from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator
    from scipy.stats._qmc import QMCEngine


class MorboTrustRegion:
    def __init__(
        self,
        num_dim: int,
        num_arms: int,
        num_metrics: int,
        *,
        rng: Generator,
    ) -> None:
        import numpy as np

        from .turbo_trust_region import TurboTrustRegion

        self._tr = TurboTrustRegion(num_dim=num_dim, num_arms=num_arms)
        self._num_dim = int(num_dim)
        self._num_arms = int(num_arms)
        self._num_metrics = int(num_metrics)
        if self._num_metrics <= 0:
            raise ValueError(self._num_metrics)

        alpha = np.ones(self._num_metrics, dtype=float)
        self._weights = np.asarray(rng.dirichlet(alpha), dtype=float)
        self._alpha = 0.05

        self._y_min: np.ndarray | Any | None = None
        self._y_max: np.ndarray | Any | None = None

    @property
    def num_dim(self) -> int:
        return self._num_dim

    @property
    def num_arms(self) -> int:
        return self._num_arms

    @property
    def num_metrics(self) -> int:
        return self._num_metrics

    @property
    def weights(self) -> np.ndarray:
        return self._weights

    @property
    def length(self) -> float:
        return float(self._tr.length)

    def _update_ranges(self, y_obs, prev_n):
        y_min_all, y_max_all = y_obs.min(axis=0), y_obs.max(axis=0)
        y_min_prev = y_obs[:prev_n].min(axis=0) if prev_n > 0 else y_min_all
        y_max_prev = y_obs[:prev_n].max(axis=0) if prev_n > 0 else y_max_all
        self._y_min, self._y_max = y_min_all, y_max_all
        return y_min_prev, y_max_prev

    def update(self, y_obs: np.ndarray | Any) -> None:
        import numpy as np

        y_obs = np.asarray(y_obs, dtype=float)
        if y_obs.ndim != 2 or y_obs.shape[1] != self._num_metrics:
            raise ValueError((y_obs.shape, self._num_metrics))
        n = int(y_obs.shape[0])
        if n == 0:
            self._y_min, self._y_max = None, None
            self._tr.restart()
            return
        prev_n = int(self._tr.prev_num_obs)
        if n < prev_n:
            raise ValueError((n, prev_n))
        y_min_prev, y_max_prev = self._update_ranges(y_obs, prev_n)
        if prev_n == 0:
            values = np.asarray(self.scalarize(y_obs, clip=True), dtype=float)
            if values.shape != (n,):
                raise RuntimeError((values.shape, n))
            self._tr.update(values)
            return
        if not np.isfinite(self._tr.best_value):
            raise RuntimeError(self._tr.best_value)
        values_old = np.asarray(
            self._scalarize_with_ranges(
                y_obs, y_min=y_min_prev, y_max=y_max_prev, clip=True
            ),
            dtype=float,
        )
        if values_old.shape != (n,):
            raise RuntimeError((values_old.shape, n))
        self._tr.best_value = float(np.max(values_old[:prev_n]))
        if prev_n < n:
            self._tr.update(values_old)

    def scalarize(self, y: np.ndarray | Any, *, clip: bool) -> np.ndarray:
        import numpy as np

        y = np.asarray(y, dtype=float)
        if y.ndim != 2 or y.shape[1] != self._num_metrics:
            raise ValueError(y.shape)
        if self._y_min is None or self._y_max is None:
            raise RuntimeError("scalarize called before any observations")

        return self._scalarize_with_ranges(
            y, y_min=self._y_min, y_max=self._y_max, clip=clip
        )

    def _scalarize_with_ranges(
        self,
        y: np.ndarray | Any,
        *,
        y_min: np.ndarray,
        y_max: np.ndarray,
        clip: bool,
    ) -> np.ndarray:
        import numpy as np

        y = np.asarray(y, dtype=float)
        if y.ndim != 2 or y.shape[1] != self._num_metrics:
            raise ValueError(y.shape)
        y_min = np.asarray(y_min, dtype=float).reshape(-1)
        y_max = np.asarray(y_max, dtype=float).reshape(-1)
        if y_min.shape != (self._num_metrics,) or y_max.shape != (self._num_metrics,):
            raise ValueError((y_min.shape, y_max.shape, self._num_metrics))

        denom = y_max - y_min
        is_deg = denom <= 0.0
        denom_safe = np.where(is_deg, 1.0, denom)
        z = (y - y_min.reshape(1, -1)) / denom_safe.reshape(1, -1)
        z = np.where(is_deg, 0.5, z)
        if clip:
            z = np.clip(z, 0.0, 1.0)
        t = z * self._weights.reshape(1, -1)
        scores = np.min(t, axis=1) + self._alpha * np.sum(t, axis=1)
        return scores

    def needs_restart(self) -> bool:
        return self._tr.needs_restart()

    def restart(self) -> None:
        self._y_min = None
        self._y_max = None
        self._tr.restart()

    def validate_request(self, num_arms: int, *, is_fallback: bool = False) -> None:
        return self._tr.validate_request(num_arms, is_fallback=is_fallback)

    def compute_bounds_1d(
        self, x_center: np.ndarray | Any, lengthscales: np.ndarray | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        return self._tr.compute_bounds_1d(x_center, lengthscales)

    def generate_candidates(
        self,
        x_center: np.ndarray,
        lengthscales: np.ndarray | None,
        num_candidates: int,
        rng: Generator,
        sobol_engine: QMCEngine,
    ) -> np.ndarray:
        from .tr_helpers import generate_tr_candidates

        return generate_tr_candidates(
            self._tr.compute_bounds_1d,
            x_center,
            lengthscales,
            num_candidates,
            rng=rng,
            sobol_engine=sobol_engine,
        )

    def get_incumbent_indices(
        self,
        y: np.ndarray | Any,
        rng: Generator,  # noqa: ARG002
    ) -> np.ndarray:
        import numpy as np

        y = np.asarray(y, dtype=float)
        if y.ndim != 2:
            raise ValueError(y.shape)
        n = y.shape[0]
        if n == 0:
            return np.array([], dtype=int)

        if y.shape[1] == 2:
            from enn.enn.enn_util import pareto_front_2d_maximize

            return pareto_front_2d_maximize(y[:, 0], y[:, 1])

        dominated = np.zeros(n, dtype=bool)
        for i in range(n):
            if dominated[i]:
                continue
            for j in range(n):
                if i == j or dominated[j]:
                    continue
                if np.all(y[j] >= y[i]) and np.any(y[j] > y[i]):
                    dominated[i] = True
                    break
        return np.where(~dominated)[0]



from __future__ import annotations

import numpy as np
import pytest

from enn.enn.enn_util import (
    calculate_sobol_indices,
    arms_from_pareto_fronts,
    pareto_front_2d_maximize,
    standardize_y,
)


def _make_sobol_synth_data(*, rng, n: int, d: int, y_2d: bool) -> tuple:
    x = rng.standard_normal((n, d))
    y = x[:, 0] + 0.1 * rng.standard_normal(n)
    if y_2d:
        y = y.reshape(-1, 1)
    return x, y


@pytest.mark.parametrize(
    "n,d,y_2d,expected_check",
    [
        (50, 3, False, lambda S: S[0] > S[1] and S[0] > S[2]),  # basic
        (50, 3, True, lambda S: np.all(S >= 0) and np.all(S <= 1)),  # y_2d
    ],
)
def test_calculate_sobol_indices_with_data(n, d, y_2d, expected_check):
    rng = np.random.default_rng(42)
    x, y = _make_sobol_synth_data(rng=rng, n=n, d=d, y_2d=y_2d)
    S = calculate_sobol_indices(x, y)
    assert S.shape == (d,) and np.all(S >= 0) and np.all(S <= 1) and expected_check(S)


@pytest.mark.parametrize(
    "make_data,expected_check",
    [
        (
            lambda rng: (rng.standard_normal((5, 2)), rng.standard_normal(5)),
            lambda S: np.all(S == 1.0),
        ),  # small_n
        (
            lambda rng: (rng.standard_normal((50, 3)), np.ones(50)),
            lambda S: np.all(S == 1.0),
        ),  # zero_variance
    ],
)
def test_calculate_sobol_indices_edge_cases(make_data, expected_check):
    rng = np.random.default_rng(42)
    x, y = make_data(rng)
    S = calculate_sobol_indices(x, y)
    d = x.shape[1]
    assert S.shape == (d,) and expected_check(S)


def test_calculate_sobol_indices_low_variance_dimension():
    rng = np.random.default_rng(42)
    n, d = 50, 3
    x = np.zeros((n, d))
    x[:, 0], x[:, 1], x[:, 2] = (
        rng.standard_normal(n),
        1e-15 * rng.standard_normal(n),
        rng.standard_normal(n),
    )
    y = x[:, 0] + x[:, 2] + 0.1 * rng.standard_normal(n)
    S = calculate_sobol_indices(x, y)
    assert S.shape == (d,) and S[1] == 0.0 and S[0] > 0 and S[2] > 0


def test_calculate_sobol_indices_dtype_preservation():
    rng = np.random.default_rng(42)
    n, d = 50, 3
    x, y = (
        rng.standard_normal((n, d)).astype(np.float32),
        rng.standard_normal(n).astype(np.float32),
    )
    S = calculate_sobol_indices(x, y)
    assert S.shape == (d,) and S.dtype == np.float32


def test_arms_from_pareto_fronts_selects_fronts_in_order():
    x_cand = np.arange(12, dtype=float).reshape(6, 2)
    mu, se = (
        np.array([5.0, 4.0, 3.0, 2.0, 1.0, 0.0]),
        np.array([0.10, 0.20, 0.15, 0.40, 0.05, 0.50]),
    )
    rng = np.random.default_rng(0)
    out4 = arms_from_pareto_fronts(x_cand, mu, se, num_arms=4, rng=rng)
    assert out4.shape == (4, 2) and np.allclose(out4, x_cand[[0, 1, 3, 5]])
    rng = np.random.default_rng(0)
    out5 = arms_from_pareto_fronts(x_cand, mu, se, num_arms=5, rng=rng)
    assert out5.shape == (5, 2) and np.allclose(out5, x_cand[[0, 1, 2, 3, 5]])


def test_pareto_front_2d_maximize_basic():
    a, b = np.array([1.0, 0.5, 0.2]), np.array([0.5, 1.0, 0.2])
    idx = pareto_front_2d_maximize(a, b)
    assert set(idx.tolist()) == {0, 1}


@pytest.mark.parametrize(
    "y_input,expected_center,check_scale",
    [
        (
            np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
            3.0,
            lambda s, y: np.isclose(s, np.std(y)),
        ),
        ([10.0, 20.0, 30.0], 20.0, lambda s, y: s > 0),
        (np.array([5.0, 5.0, 5.0, 5.0]), 5.0, lambda s, y: s == 1.0),
        ([42.0], 42.0, lambda s, y: s == 1.0),
    ],
)
def test_standardize_y(y_input, expected_center, check_scale):
    center, scale = standardize_y(y_input)
    assert center == expected_center and check_scale(scale, y_input)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator


def create_trust_region(
    config: Any,
    num_dim: int,
    num_arms: int,
    rng: Generator,
    num_metrics: int | None = None,
) -> Any:
    if config.tr_type == "none":
        from .no_trust_region import NoTrustRegion

        return NoTrustRegion(num_dim=num_dim, num_arms=num_arms)
    elif config.tr_type == "turbo":
        from .turbo_trust_region import TurboTrustRegion

        return TurboTrustRegion(num_dim=num_dim, num_arms=num_arms)
    elif config.tr_type == "morbo":
        from .morbo_trust_region import MorboTrustRegion

        effective_num_metrics = num_metrics or config.num_metrics
        if effective_num_metrics is None:
            raise ValueError("num_metrics required for tr_type='morbo'")
        return MorboTrustRegion(
            num_dim=num_dim,
            num_arms=num_arms,
            num_metrics=effective_num_metrics,
            rng=rng,
        )
    else:
        raise ValueError(f"Unknown tr_type: {config.tr_type!r}")


def get_x_center_fallback(
    config: Any,
    x_obs_list: list,
    y_obs_list: list,
    rng: Generator,
    tr_state: Any = None,
) -> np.ndarray | None:
    import numpy as np

    from .turbo_utils import argmax_random_tie

    y_array = np.asarray(y_obs_list, dtype=float)
    if y_array.size == 0:
        return None
    x_array = np.asarray(x_obs_list, dtype=float)

    if config.tr_type == "morbo" and tr_state is not None:
        if y_array.ndim == 1:
            y_array = y_array.reshape(-1, tr_state.num_metrics)
        scalarized = tr_state.scalarize(y_array, clip=True)
        idx = argmax_random_tie(scalarized, rng=rng)
    else:
        idx = argmax_random_tie(y_array, rng=rng)

    return x_array[idx]


def handle_restart_clear_always(
    x_obs_list: list,
    y_obs_list: list,
    yvar_obs_list: list,
) -> tuple[bool, int]:
    """Clear all observation lists and return (True, 0) - used by turbo_one and turbo_enn."""
    x_obs_list.clear()
    y_obs_list.clear()
    yvar_obs_list.clear()
    return True, 0


def handle_restart_check_morbo(
    config: Any,
    x_obs_list: list,
    y_obs_list: list,
    yvar_obs_list: list,
    init_idx: int,
) -> tuple[bool, int]:
    """Clear only for morbo, else preserve init_idx - used by turbo_zero and lhd_only."""
    if config.tr_type == "morbo":
        x_obs_list.clear()
        y_obs_list.clear()
        yvar_obs_list.clear()
        return True, 0
    return False, init_idx


def estimate_y_passthrough(y_observed: np.ndarray) -> np.ndarray:
    import numpy as np

    y = np.asarray(y_observed, dtype=float)
    if y.ndim == 1:
        return y.reshape(-1, 1)
    return y



from __future__ import annotations

import numpy as np

from enn.enn import EpistemicNearestNeighbors
from enn.enn.enn_params import ENNParams, PosteriorFlags


def test_conditional_posterior_function_draw_matches_unconditional_when_no_whatifs():
    rng = np.random.default_rng(0)
    train_x = rng.standard_normal((20, 3))
    train_y = train_x.sum(axis=1, keepdims=True)
    model = EpistemicNearestNeighbors(train_x, train_y, 0.1 * np.ones_like(train_y))

    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    flags = PosteriorFlags(exclude_nearest=False, observation_noise=True)
    seeds = [1, 2, 3]

    draws = model.posterior_function_draw(
        x_test, params, function_seeds=seeds, flags=flags
    )
    draws_cond = model.conditional_posterior_function_draw(
        np.zeros((0, 3), dtype=float),
        np.zeros((0, 1), dtype=float),
        x_test,
        params=params,
        function_seeds=seeds,
        flags=flags,
    )
    assert np.allclose(draws, draws_cond)


def test_conditional_posterior_function_draw_is_deterministic_and_does_not_mutate_index():
    d = 3
    train_x = np.array([[10.0, 0.0, 0.0], [0.0, 10.0, 0.0]], dtype=float)
    train_y = np.array([[0.0], [0.0]], dtype=float)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar=None)

    x_test = np.zeros((1, d), dtype=float)
    x_whatif = np.zeros((1, d), dtype=float)
    y_whatif = np.array([[5.0]], dtype=float)

    neighbors_before = model.neighbors(x_test, k=2, exclude_nearest=False)

    params = ENNParams(k=1, epi_var_scale=1.0, ale_homoscedastic_scale=1.0)
    flags = PosteriorFlags(exclude_nearest=False, observation_noise=True)
    seeds = [123, 124]

    draws1 = model.conditional_posterior_function_draw(
        x_whatif,
        y_whatif,
        x_test,
        params=params,
        function_seeds=seeds,
        flags=flags,
    )
    draws2 = model.conditional_posterior_function_draw(
        x_whatif,
        y_whatif,
        x_test,
        params=params,
        function_seeds=seeds,
        flags=flags,
    )

    neighbors_after = model.neighbors(x_test, k=2, exclude_nearest=False)
    assert len(neighbors_before) == len(neighbors_after)
    for (xb, yb), (xa, ya) in zip(neighbors_before, neighbors_after):
        assert np.allclose(xb, xa)
        assert np.allclose(yb, ya)

    assert draws1.shape == (len(seeds), 1, 1)
    assert np.allclose(draws1, draws2)


def test_conditional_posterior_function_draw_changes_with_whatifs_and_exclude_nearest():
    d = 3
    train_x = np.array([[10.0, 0.0, 0.0], [0.0, 10.0, 0.0]], dtype=float)
    train_y = np.array([[1.0], [2.0]], dtype=float)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar=None)

    x_test = np.zeros((1, d), dtype=float)
    x_whatif = np.zeros((1, d), dtype=float)
    y_whatif = np.array([[100.0]], dtype=float)

    params = ENNParams(k=1, epi_var_scale=1.0, ale_homoscedastic_scale=1.0)
    seeds = [7]

    draw_base = model.posterior_function_draw(x_test, params, function_seeds=seeds)[
        0, 0, 0
    ]
    draw_incl = model.conditional_posterior_function_draw(
        x_whatif,
        y_whatif,
        x_test,
        params=params,
        function_seeds=seeds,
        flags=PosteriorFlags(exclude_nearest=False, observation_noise=True),
    )[0, 0, 0]
    draw_excl = model.conditional_posterior_function_draw(
        x_whatif,
        y_whatif,
        x_test,
        params=params,
        function_seeds=seeds,
        flags=PosteriorFlags(exclude_nearest=True, observation_noise=True),
    )[0, 0, 0]

    assert not np.allclose(draw_base, draw_incl)
    assert not np.allclose(draw_incl, draw_excl)


def test_conditional_posterior_function_draw_matches_augmented_model_exactly():
    rng = np.random.default_rng(0)
    n_train, n_whatif, d = 8, 3, 3
    train_x = rng.standard_normal((n_train, d))
    train_y = train_x.sum(axis=1, keepdims=True).astype(float)

    x_whatif = rng.standard_normal((n_whatif, d))
    y_whatif = x_whatif.sum(axis=1, keepdims=True).astype(float)

    x_test = rng.standard_normal((5, d))
    params = ENNParams(k=4, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    flags = PosteriorFlags(exclude_nearest=True, observation_noise=True)
    function_seeds = [11, 12, 13]

    enn_a = EpistemicNearestNeighbors(train_x, train_y, train_yvar=None)
    draws_a = enn_a.conditional_posterior_function_draw(
        x_whatif,
        y_whatif,
        x_test,
        params=params,
        function_seeds=function_seeds,
        flags=flags,
    )

    enn_b = EpistemicNearestNeighbors(
        np.concatenate([train_x, x_whatif], axis=0),
        np.concatenate([train_y, y_whatif], axis=0),
        train_yvar=None,
    )
    draws_b = enn_b.posterior_function_draw(
        x_test, params, function_seeds=function_seeds, flags=flags
    )

    assert np.allclose(draws_a, draws_b)



from __future__ import annotations

import numpy as np
import pytest

from enn.turbo.turbo_config import (
    LHDOnlyConfig,
    TurboZeroConfig,
)
from enn.turbo.impl_helpers import create_trust_region
from enn.turbo.lhd_only_impl import LHDOnlyImpl
from enn.turbo.turbo_zero_impl import TurboZeroImpl


def test_turbo_zero_impl_init():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    assert impl._config is config


def test_turbo_zero_impl_get_x_center():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    rng = np.random.default_rng(42)
    x_obs = [[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]]
    y_obs = [1.0, 3.0, 2.0]
    x_center = impl.get_x_center(x_obs, y_obs, rng)
    assert x_center is not None and np.allclose(x_center, [0.3, 0.4])


def test_turbo_zero_impl_get_x_center_empty():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    rng = np.random.default_rng(42)
    x_center = impl.get_x_center([], [], rng)
    assert x_center is None


def test_turbo_zero_impl_needs_tr_list():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    assert not impl.needs_tr_list()


def test_create_trust_region_none():
    config = TurboZeroConfig(tr_type="none")
    rng = np.random.default_rng(42)
    tr = create_trust_region(config, 3, 4, rng)
    assert tr is not None and tr.num_dim == 3


def test_create_trust_region_turbo():
    config = TurboZeroConfig(tr_type="turbo")
    rng = np.random.default_rng(42)
    tr = create_trust_region(config, 3, 4, rng)
    assert tr is not None


def test_create_trust_region_morbo():
    config = TurboZeroConfig(tr_type="morbo", num_metrics=2)
    rng = np.random.default_rng(42)
    tr = create_trust_region(config, 3, 4, rng, num_metrics=2)
    assert tr is not None


def test_turbo_zero_impl_try_early_ask():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    result = impl.try_early_ask(
        4, [], lambda n: np.zeros((n, 2)), lambda n: np.zeros((n, 2))
    )
    assert result is None


def test_handle_restart_check_morbo_no_clear():
    from enn.turbo.impl_helpers import handle_restart_check_morbo

    config = TurboZeroConfig()
    should_reset, idx = handle_restart_check_morbo(config, [], [], [], 5)
    assert not should_reset and idx == 5


def test_handle_restart_check_morbo_clears():
    from enn.turbo.impl_helpers import handle_restart_check_morbo

    config = TurboZeroConfig(tr_type="morbo")
    x, y, yvar = [1], [2], [3]
    should_reset, idx = handle_restart_check_morbo(config, x, y, yvar, 5)
    assert should_reset and idx == 0 and len(x) == 0


def test_handle_restart_clear_always():
    from enn.turbo.impl_helpers import handle_restart_clear_always

    x, y, yvar = [1], [2], [3]
    should_reset, idx = handle_restart_clear_always(x, y, yvar)
    assert should_reset and idx == 0 and len(x) == 0


@pytest.mark.parametrize(
    "y_input,expected_shape",
    [
        (np.array([1.0, 2.0]), (2, 1)),
        (np.array([[1.0, 2.0], [3.0, 4.0]]), (2, 2)),
    ],
)
def test_turbo_zero_impl_estimate_y(y_input, expected_shape):
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    x = np.array([[0.1, 0.2], [0.3, 0.4]])
    result = impl.estimate_y(x, y_input)
    assert result.shape == expected_shape


def test_turbo_zero_impl_get_mu_sigma():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    result = impl.get_mu_sigma(np.zeros((5, 2)))
    assert result is None


def test_turbo_zero_impl_select_candidates():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    rng = np.random.default_rng(42)
    x_cand = np.random.rand(100, 3)
    result = impl.select_candidates(x_cand, 4, 3, rng, lambda x, n: x[:n], lambda x: x)
    assert result.shape == (4, 3)


def test_lhd_only_impl_init():
    config = LHDOnlyConfig()
    impl = LHDOnlyImpl(config)
    assert impl._config is config


def test_lhd_only_impl_select_candidates():
    config = LHDOnlyConfig()
    impl = LHDOnlyImpl(config)
    rng = np.random.default_rng(42)
    x_cand = np.random.rand(100, 3)
    result = impl.select_candidates(x_cand, 4, 3, rng, lambda x, n: x[:n], lambda x: x)
    assert result.shape == (4, 3)


def test_always_clears_on_restart_turbo_zero():
    config = TurboZeroConfig()
    impl = TurboZeroImpl(config)
    assert impl.always_clears_on_restart is False


def test_always_clears_on_restart_lhd_only():
    config = LHDOnlyConfig()
    impl = LHDOnlyImpl(config)
    assert impl.always_clears_on_restart is False


def test_always_clears_on_restart_turbo_one():
    from enn.turbo.turbo_config import TurboOneConfig
    from enn.turbo.turbo_one_impl import TurboOneImpl

    config = TurboOneConfig()
    impl = TurboOneImpl(config)
    assert impl.always_clears_on_restart is True


def test_always_clears_on_restart_turbo_enn():
    from enn.turbo.turbo_config import TurboENNConfig
    from enn.turbo.turbo_enn_impl import TurboENNImpl

    config = TurboENNConfig()
    impl = TurboENNImpl(config)
    assert impl.always_clears_on_restart is True


def test_estimate_y_passthrough():
    from enn.turbo.impl_helpers import estimate_y_passthrough

    y_1d = np.array([1.0, 2.0, 3.0])
    result = estimate_y_passthrough(y_1d)
    assert result.shape == (3, 1)

    y_2d = np.array([[1.0, 2.0], [3.0, 4.0]])
    result = estimate_y_passthrough(y_2d)
    assert result.shape == (2, 2)


def test_get_x_center_fallback():
    from enn.turbo.impl_helpers import get_x_center_fallback

    config = TurboZeroConfig()
    rng = np.random.default_rng(42)
    x_obs = [[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]]
    y_obs = [1.0, 3.0, 2.0]
    result = get_x_center_fallback(config, x_obs, y_obs, rng)
    assert result is not None
    assert np.allclose(result, [0.3, 0.4])


def test_turbo_optimizer_telemetry():
    from enn.turbo import TurboOptimizer, TurboMode

    rng = np.random.default_rng(42)
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]])
    opt = TurboOptimizer(bounds, TurboMode.TURBO_ZERO, rng=rng)
    tel = opt.telemetry()
    assert tel.dt_fit == 0.0
    assert tel.dt_sel == 0.0


def test_generate_raasp_candidates():
    from enn.turbo.turbo_utils import generate_raasp_candidates
    from scipy.stats import qmc

    rng = np.random.default_rng(42)
    center = np.array([0.5, 0.5, 0.5])
    lb = np.array([0.0, 0.0, 0.0])
    ub = np.array([1.0, 1.0, 1.0])
    sobol = qmc.Sobol(d=3, scramble=True, seed=42)
    result = generate_raasp_candidates(center, lb, ub, 100, rng=rng, sobol_engine=sobol)
    assert result.shape == (100, 3)
    assert np.all(result >= 0.0) and np.all(result <= 1.0)


def _assert_candidates_in_unit_bounds(fn):
    rng = np.random.default_rng(42)
    center = np.array([0.5, 0.5, 0.5])
    lb = np.array([0.0, 0.0, 0.0])
    ub = np.array([1.0, 1.0, 1.0])
    result = fn(center, lb, ub, 100, rng=rng)
    assert result.shape == (100, 3)
    assert np.all(result >= 0.0) and np.all(result <= 1.0)


def test_generate_raasp_candidates_uniform():
    from enn.turbo.turbo_utils import generate_raasp_candidates_uniform

    _assert_candidates_in_unit_bounds(generate_raasp_candidates_uniform)



from __future__ import annotations

import numpy as np

from enn.enn import EpistemicNearestNeighbors
from enn.enn.enn_params import ENNParams, PosteriorFlags


def test_posterior_function_sample_basic():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    train_y = train_x.sum(axis=1, keepdims=True)
    model = EpistemicNearestNeighbors(train_x, train_y, 0.1 * np.ones_like(train_y))
    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    sample = model.posterior_function_draw(x_test, params, function_seeds=[123])[0]
    assert sample.shape == (5, 1) and np.all(np.isfinite(sample))


def test_posterior_function_sample_deterministic():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    model = EpistemicNearestNeighbors(train_x, train_x.sum(axis=1, keepdims=True))
    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    sample1 = model.posterior_function_draw(x_test, params, function_seeds=[42])[0]
    assert np.allclose(
        sample1, model.posterior_function_draw(x_test, params, function_seeds=[42])[0]
    )
    assert not np.allclose(
        sample1, model.posterior_function_draw(x_test, params, function_seeds=[43])[0]
    )


def test_posterior_function_sample_batch_basic():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    train_y = train_x.sum(axis=1, keepdims=True)
    model = EpistemicNearestNeighbors(train_x, train_y, 0.1 * np.ones_like(train_y))
    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    samples = model.posterior_function_draw(x_test, params, function_seeds=[10, 20, 30])
    assert samples.shape == (3, 5, 1) and np.all(np.isfinite(samples))


def test_posterior_function_sample_batch_matches_single_seed():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    model = EpistemicNearestNeighbors(train_x, train_x.sum(axis=1, keepdims=True))
    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    batch = model.posterior_function_draw(
        x_test, params, function_seeds=[100, 200, 300]
    )
    for i, seed in enumerate([100, 200, 300]):
        assert np.allclose(
            batch[i],
            model.posterior_function_draw(x_test, params, function_seeds=[seed])[0],
        )


def test_posterior_function_sample_batch_with_multiple_metrics():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    model = EpistemicNearestNeighbors(train_x, rng.standard_normal((20, 2)))
    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    samples = model.posterior_function_draw(x_test, params, function_seeds=[1, 2, 3, 4])
    assert samples.shape == (4, 5, 2) and np.all(np.isfinite(samples))


def test_posterior_function_sample_batch_empty_k():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((2, 3))
    train_y = train_x.sum(axis=1, keepdims=True)
    model = EpistemicNearestNeighbors(train_x, train_y)
    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    samples = model.posterior_function_draw(
        x_test,
        params,
        function_seeds=[1, 2],
        flags=PosteriorFlags(exclude_nearest=True),
    )
    assert samples.shape == (2, 5, 1)


def test_posterior_function_sample_with_observation_noise():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    train_y = train_x.sum(axis=1, keepdims=True)
    model = EpistemicNearestNeighbors(train_x, train_y, 0.5 * np.ones_like(train_y))
    x_test = rng.standard_normal((5, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    sample_no_noise = model.posterior_function_draw(
        x_test, params, function_seeds=[42]
    )[0]
    sample_with_noise = model.posterior_function_draw(
        x_test,
        params,
        function_seeds=[42],
        flags=PosteriorFlags(observation_noise=True),
    )[0]
    assert sample_no_noise.shape == sample_with_noise.shape == (5, 1)
    assert not np.allclose(sample_no_noise, sample_with_noise)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np


def sobol_seed_for_state(seed_base: int, *, n_obs: int, num_arms: int) -> int:
    mask64 = (1 << 64) - 1
    x = int(seed_base) & mask64
    x ^= (int(n_obs) + 1) * 0x9E3779B97F4A7C15 & mask64
    x ^= (int(num_arms) + 1) * 0xBF58476D1CE4E5B9 & mask64
    x = (x + 0x9E3779B97F4A7C15) & mask64
    z = x
    z = (z ^ (z >> 30)) * 0xBF58476D1CE4E5B9 & mask64
    z = (z ^ (z >> 27)) * 0x94D049BB133111EB & mask64
    z = z ^ (z >> 31)
    return int(z & 0xFFFFFFFF)


def validate_tell_inputs(
    x: np.ndarray, y: np.ndarray, y_var: np.ndarray | None, num_dim: int
) -> tuple[np.ndarray, np.ndarray, np.ndarray | None, int]:
    import numpy as np

    x = np.asarray(x, dtype=float)
    y = np.asarray(y, dtype=float)
    if x.ndim != 2 or x.shape[1] != num_dim:
        raise ValueError(x.shape)

    if y.ndim == 2:
        if y.shape[0] != x.shape[0]:
            raise ValueError((x.shape, y.shape))
        num_metrics = y.shape[1]
    elif y.ndim == 1:
        if y.shape[0] != x.shape[0]:
            raise ValueError((x.shape, y.shape))
        num_metrics = 1
    else:
        raise ValueError(y.shape)

    if y_var is not None:
        y_var = np.asarray(y_var, dtype=float)
        if y_var.shape != y.shape:
            raise ValueError((y.shape, y_var.shape))

    return x, y, y_var, num_metrics


def trim_trailing_observations(
    x_obs_list: list,
    y_obs_list: list,
    y_tr_list: list,
    yvar_obs_list: list,
    *,
    trailing_obs: int,
    incumbent_indices: np.ndarray,
) -> tuple[list, list, list, list]:
    import numpy as np

    num_total = len(x_obs_list)
    if num_total <= trailing_obs:
        return x_obs_list, y_obs_list, y_tr_list, yvar_obs_list

    start_idx = max(0, num_total - trailing_obs)
    recent_indices = set(range(start_idx, num_total))
    keep_indices = set(incumbent_indices.tolist()) | recent_indices

    if len(keep_indices) > trailing_obs:
        keep_indices = set(incumbent_indices.tolist())
        remaining_slots = trailing_obs - len(keep_indices)
        if remaining_slots > 0:
            recent_non_incumbent = [
                i for i in range(num_total - 1, -1, -1) if i not in keep_indices
            ][:remaining_slots]
            keep_indices.update(recent_non_incumbent)

    indices = np.array(sorted(keep_indices), dtype=int)

    x_array = np.asarray(x_obs_list, dtype=float)
    y_obs_array = np.asarray(y_obs_list, dtype=float)
    y_tr_array = np.asarray(y_tr_list, dtype=float)

    new_x = x_array[indices].tolist()
    new_y_obs = y_obs_array[indices].tolist()
    new_y_tr = y_tr_array[indices].tolist()
    new_yvar = yvar_obs_list
    if len(yvar_obs_list) == len(y_obs_array):
        yvar_array = np.asarray(yvar_obs_list, dtype=float)
        new_yvar = yvar_array[indices].tolist()

    return new_x, new_y_obs, new_y_tr, new_yvar



from __future__ import annotations

import sys
from pathlib import Path

# Import torch before faiss to avoid OpenMP conflict on macOS.
# Both libraries use OpenMP; torch must load its runtime first.
import torch  # noqa: F401

src_path = Path(__file__).parent.parent / "src"
if str(src_path) not in sys.path:
    sys.path.insert(0, str(src_path))


def sphere_objective(x):
    import numpy as np

    return -np.sum(x**2, axis=1)


def make_from_unit_fn(bounds):
    from enn.turbo.turbo_utils import from_unit

    def from_unit_fn(x):
        return from_unit(x, bounds)

    return from_unit_fn


def make_select_sobol_fn(bounds, rng):
    from enn.turbo.turbo_utils import from_unit

    def select_sobol_fn(x, n):
        idx = rng.choice(x.shape[0], size=n, replace=False)
        return from_unit(x[idx], bounds)

    return select_sobol_fn


def make_enn_model(n=20, d=3, seed=0, yvar_scale=0.1):
    import numpy as np

    from enn.enn import EpistemicNearestNeighbors

    rng = np.random.default_rng(seed)
    train_x = rng.standard_normal((n, d))
    train_y = (train_x.sum(axis=1, keepdims=True)).astype(float)
    train_yvar = yvar_scale * np.ones_like(train_y)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    return model, train_x, train_y, train_yvar, rng



from __future__ import annotations

from .enn import EpistemicNearestNeighbors, enn_fit

_LAZY_IMPORTS = ("TurboMode", "TurboOptimizer", "Turbo", "Telemetry")


def _lazy_load(name: str):
    from . import turbo

    return getattr(turbo, name)


def __getattr__(name: str):
    if name in _LAZY_IMPORTS:
        return _lazy_load(name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__: list[str] = [
    "EpistemicNearestNeighbors",
    "enn_fit",
    *_LAZY_IMPORTS,
]



from __future__ import annotations

import numpy as np
import pytest

from enn.enn import EpistemicNearestNeighbors
from enn.enn.enn_params import ENNParams, PosteriorFlags


def _params(
    k: int, *, epi_var_scale: float = 1.0, ale_homoscedastic_scale: float = 0.0
):
    return ENNParams(
        k=int(k),
        epi_var_scale=float(epi_var_scale),
        ale_homoscedastic_scale=float(ale_homoscedastic_scale),
    )


def test_ennnormal_sample_shape_and_clip():
    from enn.enn.enn_normal import ENNNormal

    rng = np.random.default_rng(0)
    mu = np.array([[0.0, 1.0]], dtype=float)
    se = np.array([[1.0, 2.0]], dtype=float)
    normal = ENNNormal(mu=mu, se=se)
    samples = normal.sample(5, clip=1.0, rng=rng)
    assert samples.shape == (1, 2, 5)
    assert np.all(samples >= mu.min() - 2.0)
    assert np.all(samples <= mu.max() + 2.0)


def test_epistemic_nearest_neighbors_posterior_and_var_scale():
    import conftest

    model, _train_x, _train_y, _train_yvar, rng = conftest.make_enn_model()
    x_test = rng.standard_normal((4, 3))
    params = ENNParams(k=3, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    post = model.posterior(x_test, params=params)
    assert post.mu.shape == (4, 1)
    assert post.se.shape == (4, 1)
    post_changed = model.posterior(
        x_test,
        params=ENNParams(k=5, epi_var_scale=0.5, ale_homoscedastic_scale=0.0),
        flags=PosteriorFlags(exclude_nearest=True),
    )
    assert post_changed.mu.shape == (4, 1)
    assert post_changed.se.shape == (4, 1)


def test_epistemic_nearest_neighbors_with_no_observations_returns_prior_like_posterior():
    rng = np.random.default_rng(0)
    d = 3
    x = np.zeros((0, d), dtype=float)
    y = np.zeros((0, 1), dtype=float)
    yvar = np.ones_like(y, dtype=float)
    model = EpistemicNearestNeighbors(x, y, yvar)
    x_test = rng.standard_normal((5, d))
    post = model.posterior(
        x_test,
        params=ENNParams(k=3, epi_var_scale=1.0, ale_homoscedastic_scale=0.0),
    )
    assert post.mu.shape == (5, 1)
    assert post.se.shape == (5, 1)
    assert np.allclose(post.mu, 0.0)
    assert np.allclose(post.se, 1.0)


@pytest.mark.parametrize("num_obs", [1, 2, 3])
def test_epistemic_nearest_neighbors_with_few_observations_has_valid_posterior(
    num_obs: int,
):
    rng = np.random.default_rng(0)
    d = 3
    x = rng.standard_normal((num_obs, d))
    y = (x.sum(axis=1, keepdims=True)).astype(float)
    yvar = 0.1 * np.ones_like(y)
    model = EpistemicNearestNeighbors(x, y, yvar)
    x_test = rng.standard_normal((5, d))
    post = model.posterior(x_test, params=_params(3))
    assert post.mu.shape == (5, 1)
    assert post.se.shape == (5, 1)
    assert np.all(np.isfinite(post.mu))
    assert np.all(np.isfinite(post.se))


@pytest.mark.parametrize(
    "flags,k_vals",
    [
        (PosteriorFlags(), [3, 5, 7]),
        (PosteriorFlags(exclude_nearest=True), [3, 5]),
    ],
)
def test_batch_posterior_matches_individual_posterior_calls(flags, k_vals):
    import conftest

    model, _train_x, _train_y, _train_yvar, rng = conftest.make_enn_model()
    x_test = rng.standard_normal((4, 3))
    paramss = [
        ENNParams(k=k, epi_var_scale=1.0 / (i + 1), ale_homoscedastic_scale=0.0)
        for i, k in enumerate(k_vals)
    ]
    post_batch = model.batch_posterior(x_test, paramss, flags=flags)
    assert post_batch.mu.shape == (len(paramss), x_test.shape[0], model.num_outputs)
    assert post_batch.se.shape == (len(paramss), x_test.shape[0], model.num_outputs)
    for i, params in enumerate(paramss):
        post = model.posterior(x_test, params=params, flags=flags)
        assert np.allclose(post_batch.mu[i], post.mu) and np.allclose(
            post_batch.se[i], post.se
        )


def test_epistemic_nearest_neighbors_with_sobol_indices():
    rng = np.random.default_rng(0)
    n = 50
    d = 3
    x = rng.standard_normal((n, d))
    y = (x[:, 0] + 0.1 * x[:, 1] + 0.01 * rng.standard_normal(n)).reshape(-1, 1)
    yvar = 0.1 * np.ones_like(y)
    model = EpistemicNearestNeighbors(x, y, yvar)
    x_test = rng.standard_normal((4, d))
    params = ENNParams(k=3, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    post = model.posterior(x_test, params=params)
    assert post.mu.shape == (4, 1) and post.se.shape == (4, 1)
    assert np.all(np.isfinite(post.mu)) and np.all(np.isfinite(post.se))


def test_epistemic_nearest_neighbors_multiple_metrics():
    rng = np.random.default_rng(0)
    n = 20
    d = 3
    x = rng.standard_normal((n, d))
    y = rng.standard_normal((n, 2))
    yvar = 0.1 * np.ones_like(y)
    model = EpistemicNearestNeighbors(x, y, yvar)
    x_test = rng.standard_normal((4, d))
    params = ENNParams(k=3, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    post = model.posterior(x_test, params=params)
    assert post.mu.shape == (4, 2) and post.se.shape == (4, 2)


def test_batch_posterior_exclude_nearest_with_k_larger_than_available():
    """
    Forces the off-by-one bug when exclude_nearest=True and k > len(self) - 1.

    With len(self)=5, max_k=10, exclude_nearest=True:
    - search_k = min(11, 5) = 5
    - After slicing [:, 1:], arrays have 4 columns
    - BUG: k = min(10, 5) = 5, but should be min(10, 4) = 4
    """
    rng = np.random.default_rng(0)
    n = 5
    d = 3
    train_x = rng.standard_normal((n, d))
    train_y = (train_x.sum(axis=1, keepdims=True)).astype(float)
    train_yvar = 0.1 * np.ones_like(train_y)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)

    x_test = rng.standard_normal((4, d))
    params = ENNParams(k=10, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    post = model.batch_posterior(
        x_test, [params], flags=PosteriorFlags(exclude_nearest=True)
    )
    assert post.mu.shape == (1, 4, 1) and post.se.shape == (1, 4, 1)
    assert np.all(np.isfinite(post.mu))
    assert np.all(np.isfinite(post.se))


def test_epistemic_nearest_neighbors_with_yvar_none():
    rng = np.random.default_rng(42)
    n = 20
    d = 3
    train_x = rng.standard_normal((n, d))
    train_y = train_x.sum(axis=1, keepdims=True) + rng.standard_normal((n, 1)) * 0.1

    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar=None)

    assert len(model) == n
    assert model.train_yvar is None

    x_test = rng.standard_normal((10, d))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    post = model.posterior(x_test, params=params)

    assert post.mu.shape == (10, 1)
    assert post.se.shape == (10, 1)
    assert np.all(np.isfinite(post.mu))
    assert np.all(np.isfinite(post.se))


def test_epistemic_nearest_neighbors_constant_y_scale_is_safe():
    rng = np.random.default_rng(0)
    n = 20
    d = 3
    train_x = rng.standard_normal((n, d))
    train_y = np.zeros((n, 1), dtype=float)
    train_yvar = 0.1 * np.ones_like(train_y)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)

    x_test = rng.standard_normal((5, d))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    post = model.posterior(x_test, params=params)
    assert np.all(np.isfinite(post.mu))
    assert np.all(np.isfinite(post.se))


def test_epistemic_nearest_neighbors_init_validates_inputs():
    rng = np.random.default_rng(0)

    # x must be 2D
    with pytest.raises(ValueError):
        EpistemicNearestNeighbors(rng.random(10), np.zeros((10, 1)))

    # y must be 2D
    with pytest.raises(ValueError):
        EpistemicNearestNeighbors(rng.random((10, 3)), rng.random(10))

    # x and y must have matching rows
    with pytest.raises(ValueError):
        EpistemicNearestNeighbors(rng.random((10, 3)), rng.random((5, 1)))

    # yvar must be 2D if provided
    with pytest.raises(ValueError):
        EpistemicNearestNeighbors(
            rng.random((10, 3)), rng.random((10, 1)), rng.random(10)
        )

    # yvar must match y shape
    with pytest.raises(ValueError):
        EpistemicNearestNeighbors(
            rng.random((10, 3)), rng.random((10, 1)), rng.random((10, 2))
        )


def test_epistemic_nearest_neighbors_init_explicit():
    rng = np.random.default_rng(42)
    n, d = 20, 3
    train_x = rng.standard_normal((n, d))
    train_y = train_x.sum(axis=1, keepdims=True)
    train_yvar = 0.1 * np.ones_like(train_y)

    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    assert len(model) == n
    assert model.num_outputs == 1
    assert model.train_x is not None
    assert model.train_y is not None
    assert model.train_yvar is not None



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

import numpy as np

if TYPE_CHECKING:
    from numpy.random import Generator

from .turbo_config import TurboOneConfig
from .turbo_utils import get_gp_posterior_suppress_warning, gp_thompson_sample


class TurboOneImpl:
    def __init__(self, config: TurboOneConfig) -> None:
        self._config = config
        self._gp_model: Any | None = None
        self._gp_y_mean: float | Any = 0.0
        self._gp_y_std: float | Any = 1.0

    @property
    def always_clears_on_restart(self) -> bool:
        return True
        self._fitted_n_obs: int = 0

    def _as_2d(self, a: np.ndarray) -> np.ndarray:
        a = np.asarray(a, dtype=float)
        if a.ndim == 1:
            return a.reshape(-1, 1)
        if a.ndim == 2:
            return a.T
        raise ValueError(a.shape)

    def _broadcast_gp_mean_std(self, num_metrics: int) -> tuple[np.ndarray, np.ndarray]:
        num_metrics = int(num_metrics)
        if num_metrics <= 0:
            raise ValueError(num_metrics)
        gp_y_mean = np.asarray(self._gp_y_mean, dtype=float).reshape(-1)
        gp_y_std = np.asarray(self._gp_y_std, dtype=float).reshape(-1)
        if gp_y_mean.size == 1 and num_metrics != 1:
            gp_y_mean = np.full(num_metrics, float(gp_y_mean[0]), dtype=float)
        if gp_y_std.size == 1 and num_metrics != 1:
            gp_y_std = np.full(num_metrics, float(gp_y_std[0]), dtype=float)
        if gp_y_mean.shape != (num_metrics,) or gp_y_std.shape != (num_metrics,):
            raise ValueError((gp_y_mean.shape, gp_y_std.shape, num_metrics))
        return gp_y_mean, gp_y_std

    def _unstandardize(self, y_std_2d: np.ndarray) -> np.ndarray:
        y_std_2d = np.asarray(y_std_2d, dtype=float)
        if y_std_2d.ndim != 2:
            raise ValueError(y_std_2d.shape)
        num_metrics = int(y_std_2d.shape[1])
        gp_y_mean, gp_y_std = self._broadcast_gp_mean_std(num_metrics)
        return gp_y_mean.reshape(1, -1) + gp_y_std.reshape(1, -1) * y_std_2d

    def _score_y_2d(self, y_array: np.ndarray, tr_state: Any | None) -> np.ndarray:
        if self._config.tr_type == "morbo" and tr_state is not None:
            return tr_state.scalarize(y_array, clip=True)
        return y_array[:, 0]

    def _get_x_center_no_gp(
        self,
        x_obs_list: list,
        y_obs_list: list,
        rng: Generator,
        tr_state: Any = None,
    ) -> np.ndarray | None:
        from .impl_helpers import get_x_center_fallback
        from .turbo_utils import argmax_random_tie

        x_array = np.asarray(x_obs_list, dtype=float)
        y_array = np.asarray(y_obs_list, dtype=float)
        if y_array.ndim != 2:
            return get_x_center_fallback(
                self._config, x_obs_list, y_obs_list, rng, tr_state
            )
        scores = self._score_y_2d(y_array, tr_state)
        best_idx = argmax_random_tie(scores, rng=rng)
        return x_array[best_idx]

    def get_x_center(
        self,
        x_obs_list: list,
        y_obs_list: list,
        rng: Generator,
        tr_state: Any = None,
    ) -> np.ndarray | None:
        import torch

        from .turbo_utils import argmax_random_tie

        if len(y_obs_list) == 0:
            return None
        if self._gp_model is None and len(y_obs_list) <= 1:
            return self._get_x_center_no_gp(x_obs_list, y_obs_list, rng, tr_state)
        if self._gp_model is None:
            raise RuntimeError(
                "TurboOneImpl.get_x_center requires a fitted GP model for 2+ observations; "
                "call prepare_ask() first."
            )
        if self._fitted_n_obs != len(x_obs_list):
            raise RuntimeError(
                f"GP fitted on {self._fitted_n_obs} obs but get_x_center called with {len(x_obs_list)}"
            )

        x_array = np.asarray(x_obs_list, dtype=float)
        x_torch = torch.as_tensor(x_array, dtype=torch.float64)

        with torch.no_grad():
            posterior = get_gp_posterior_suppress_warning(self._gp_model, x_torch)
            mu_std = posterior.mean.cpu().numpy()

        mu = self._unstandardize(self._as_2d(mu_std))

        if self._config.tr_type == "morbo" and tr_state is not None:
            scalarized = tr_state.scalarize(mu, clip=False)
            best_idx = argmax_random_tie(scalarized, rng=rng)
        else:
            if mu.shape[1] != 1:
                raise ValueError(mu.shape)
            best_idx = argmax_random_tie(mu[:, 0], rng=rng)

        return x_array[best_idx]

    def needs_tr_list(self) -> bool:
        return True

    def try_early_ask(
        self,
        num_arms: int,
        x_obs_list: list,
        draw_initial_fn: Callable[[int], np.ndarray],
        get_init_lhd_points_fn: Callable[[int], np.ndarray],
    ) -> np.ndarray | None:
        return get_init_lhd_points_fn(num_arms) if len(x_obs_list) == 0 else None

    def prepare_ask(
        self,
        x_obs_list: list,
        y_obs_list: list,
        yvar_obs_list: list,
        num_dim: int,
        gp_num_steps: int,
        rng: Any | None = None,
    ) -> tuple[Any, float | None, float | None, np.ndarray | None]:
        from .turbo_gp_fit import fit_gp

        if len(x_obs_list) == 0:
            return None, None, None, None
        self._gp_model, _likelihood, gp_y_mean_fitted, gp_y_std_fitted = fit_gp(
            x_obs_list,
            y_obs_list,
            num_dim,
            yvar_obs_list=yvar_obs_list if yvar_obs_list else None,
            num_steps=gp_num_steps,
        )
        self._fitted_n_obs = len(x_obs_list)
        if gp_y_mean_fitted is not None:
            self._gp_y_mean = gp_y_mean_fitted
        if gp_y_std_fitted is not None:
            self._gp_y_std = gp_y_std_fitted
        lengthscales = None
        if self._gp_model is not None:
            lengthscale = (
                self._gp_model.covar_module.base_kernel.lengthscale.cpu()
                .detach()
                .numpy()
            )
            if lengthscale.ndim == 3:
                lengthscale = lengthscale.mean(axis=0)
            lengthscales = lengthscale.ravel()
            # First line helps stabilize second line.
            lengthscales = lengthscales / lengthscales.mean()
            lengthscales = lengthscales / np.prod(
                np.power(lengthscales, 1.0 / len(lengthscales))
            )
        return self._gp_model, gp_y_mean_fitted, gp_y_std_fitted, lengthscales

    def select_candidates(
        self,
        x_cand: np.ndarray,
        num_arms: int,
        num_dim: int,
        rng: Generator,
        fallback_fn: Callable[[np.ndarray, int], np.ndarray],
        from_unit_fn: Callable[[np.ndarray], np.ndarray],
        tr_state: Any = None,
    ) -> np.ndarray:
        if self._gp_model is None:
            if self._fitted_n_obs >= 2:
                raise RuntimeError(
                    "TurboOneImpl.select_candidates requires a fitted GP model for 2+ observations; "
                    "call prepare_ask() first."
                )
            return fallback_fn(x_cand, num_arms)

        if self._config.tr_type == "morbo" and tr_state is not None:
            import gpytorch
            import torch

            from .turbo_utils import torch_seed_context

            x_torch = torch.as_tensor(x_cand, dtype=torch.float64)
            seed = int(rng.integers(2**31 - 1))
            with (
                torch.no_grad(),
                gpytorch.settings.fast_pred_var(),
                torch_seed_context(seed, device=x_torch.device),
            ):
                posterior = self._gp_model.posterior(x_torch)
                samples = posterior.sample(sample_shape=torch.Size([1]))

            if samples.ndim == 2:
                samples_std = samples[0].detach().cpu().numpy().reshape(-1, 1)
            elif samples.ndim == 3:
                samples_std = samples[0].detach().cpu().numpy().T
            else:
                raise ValueError(samples.shape)

            y_samples = self._unstandardize(samples_std)
            scores = tr_state.scalarize(y_samples, clip=False)
            shuffled_indices = rng.permutation(len(scores))
            shuffled_scores = scores[shuffled_indices]
            top_k_in_shuffled = np.argpartition(-shuffled_scores, num_arms - 1)[
                :num_arms
            ]
            idx = shuffled_indices[top_k_in_shuffled]
            return from_unit_fn(x_cand[idx])

        if (
            np.asarray(self._gp_y_mean).ndim != 0
            or np.asarray(self._gp_y_std).ndim != 0
        ):
            raise ValueError("multi-output GP requires tr_type='morbo'")
        idx = gp_thompson_sample(
            self._gp_model,
            x_cand,
            num_arms,
            rng,
            gp_y_mean=float(self._gp_y_mean),
            gp_y_std=float(self._gp_y_std),
        )
        return from_unit_fn(x_cand[idx])

    def estimate_y(self, x_unit: np.ndarray, y_observed: np.ndarray) -> np.ndarray:
        import torch

        if self._gp_model is None:
            raise RuntimeError(
                "TurboOneImpl.estimate_y requires a fitted GP model; call prepare_ask() first."
            )

        x_torch = torch.as_tensor(x_unit, dtype=torch.float64)
        with torch.no_grad():
            posterior = get_gp_posterior_suppress_warning(self._gp_model, x_torch)
            mu_std = posterior.mean.cpu().numpy()

        mu = self._unstandardize(self._as_2d(mu_std))
        return mu

    def get_mu_sigma(self, x_unit: np.ndarray) -> tuple[np.ndarray, np.ndarray] | None:
        import torch

        if self._gp_model is None:
            return None
        x_torch = torch.as_tensor(x_unit, dtype=torch.float64)
        with torch.no_grad():
            posterior = self._gp_model.posterior(x_torch)
            mu_std = posterior.mean.cpu().numpy()
            sigma_std = posterior.variance.cpu().numpy() ** 0.5

        mu_std_2d = self._as_2d(mu_std)
        sigma_std_2d = self._as_2d(sigma_std)
        mu = self._unstandardize(mu_std_2d)
        _gp_y_mean, gp_y_std = self._broadcast_gp_mean_std(int(mu_std_2d.shape[1]))
        sigma = gp_y_std.reshape(1, -1) * sigma_std_2d
        return mu, sigma



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator
    from scipy.stats._qmc import QMCEngine


def validate_trust_region_request(
    num_arms: int, configured_num_arms: int, *, is_fallback: bool = False
) -> None:
    if is_fallback:
        if num_arms > configured_num_arms:
            raise ValueError(
                f"num_arms {num_arms} > configured num_arms {configured_num_arms}"
            )
    else:
        if num_arms != configured_num_arms:
            raise ValueError(
                f"num_arms {num_arms} != configured num_arms {configured_num_arms}"
            )


def generate_tr_candidates(
    compute_bounds_1d: Any,
    x_center: np.ndarray,
    lengthscales: np.ndarray | None,
    num_candidates: int,
    *,
    rng: Generator,
    candidate_rv: Literal["sobol", "uniform"] = "sobol",
    sobol_engine: QMCEngine | None = None,
) -> np.ndarray:
    from .turbo_utils import (
        generate_raasp_candidates,
        generate_raasp_candidates_uniform,
    )

    lb, ub = compute_bounds_1d(x_center, lengthscales)
    if candidate_rv == "sobol":
        if sobol_engine is None:
            raise ValueError("sobol_engine is required when candidate_rv='sobol'")
        return generate_raasp_candidates(
            x_center, lb, ub, num_candidates, rng=rng, sobol_engine=sobol_engine
        )
    if candidate_rv == "uniform":
        return generate_raasp_candidates_uniform(
            x_center, lb, ub, num_candidates, rng=rng
        )
    raise ValueError(candidate_rv)



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class PosteriorFlags:
    exclude_nearest: bool = False
    observation_noise: bool = False


@dataclass(frozen=True)
class ENNParams:
    k: int
    epi_var_scale: float
    ale_homoscedastic_scale: float

    def __post_init__(self) -> None:
        import numpy as np

        k = int(self.k)
        if k <= 0:
            raise ValueError(f"k must be > 0, got {k}")
        epi_var_scale = float(self.epi_var_scale)
        if not np.isfinite(epi_var_scale) or epi_var_scale < 0.0:
            raise ValueError(f"epi_var_scale must be >= 0, got {epi_var_scale}")
        ale_scale = float(self.ale_homoscedastic_scale)
        if not np.isfinite(ale_scale) or ale_scale < 0.0:
            raise ValueError(f"ale_homoscedastic_scale must be >= 0, got {ale_scale}")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from gpytorch.distributions import MultivariateNormal


def _get_exact_gp_base():
    from gpytorch.models import ExactGP

    return ExactGP


class TurboGPBase(_get_exact_gp_base()):
    mean_module: Any
    covar_module: Any

    def forward(self, x) -> MultivariateNormal:
        from gpytorch.distributions import MultivariateNormal

        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)

    def posterior(self, x) -> MultivariateNormal:
        return self(x)



"""Performance regression test for ENN demo operations."""

from __future__ import annotations

import time

import numpy as np
import pytest

from enn import EpistemicNearestNeighbors, enn_fit
from enn.enn.enn_params import PosteriorFlags


def make_enn_demo_data(num_samples: int, k: int, noise: float, m: int = 1):
    x = np.sort(np.random.rand(num_samples + 4))
    x[-3] = x[-4]
    x[-2] = x[-4]
    x[-1] = x[-4]
    x[1] = x[0] + 0.03
    eps = np.random.randn(num_samples + 4)
    y = np.sin(2 * m * np.pi * x) + noise * eps
    yvar = (noise**2) * np.ones_like(y)
    model = EpistemicNearestNeighbors(x[:, None], y[:, None], yvar[:, None])
    rng = np.random.default_rng(0)
    params = enn_fit(
        model,
        k=k,
        num_fit_candidates=100,
        num_fit_samples=min(10, num_samples),
        rng=rng,
    )
    return x, y, model, params


def plot_enn_posterior_logic(model, params):
    x_hat = np.linspace(0.0, 1.0, 30)
    x_hat_2d = x_hat[:, None]
    posterior = model.posterior(
        x_hat_2d, params=params, flags=PosteriorFlags(exclude_nearest=False)
    )
    return posterior.mu[:, 0], posterior.se[:, 0]


@pytest.mark.skip(reason="Timing unreliable across machines; run manually to verify")
def test_enn_demo_performance():
    """Ensure demo_enn.ipynb cell 3 runs in under 0.3 seconds."""
    np.random.seed(1)

    t0 = time.time()
    x, y, model, params = make_enn_demo_data(num_samples=1_000_000, k=5, noise=0.3, m=3)
    mu, se = plot_enn_posterior_logic(model, params)
    elapsed = time.time() - t0

    print(f"\nTime taken: {elapsed:.3f} seconds")

    assert elapsed < 0.3, f"Expected < 0.3s, got {elapsed:.3f}s"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Protocol

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator


class TurboModeImpl(Protocol):
    @property
    def always_clears_on_restart(self) -> bool: ...

    def get_x_center(
        self,
        x_obs_list: list,
        y_obs_list: list,
        rng: Generator,
        tr_state: Any = None,
    ) -> np.ndarray | None: ...

    def needs_tr_list(self) -> bool: ...

    def try_early_ask(
        self,
        num_arms: int,
        x_obs_list: list,
        draw_initial_fn: Callable[[int], np.ndarray],
        get_init_lhd_points_fn: Callable[[int], np.ndarray],
    ) -> np.ndarray | None: ...

    def prepare_ask(
        self,
        x_obs_list: list,
        y_obs_list: list,
        yvar_obs_list: list,
        num_dim: int,
        gp_num_steps: int,
        rng: Generator | Any | None = None,
    ) -> tuple[Any, float | None, float | None, np.ndarray | None]: ...

    def select_candidates(
        self,
        x_cand: np.ndarray,
        num_arms: int,
        num_dim: int,
        rng: Generator,
        fallback_fn: Callable[[np.ndarray, int], np.ndarray],
        from_unit_fn: Callable[[np.ndarray], np.ndarray],
        tr_state: Any = None,
    ) -> np.ndarray: ...

    def estimate_y(self, x_unit: np.ndarray, y_observed: np.ndarray) -> np.ndarray: ...



from __future__ import annotations

import numpy as np
import pytest

from enn.enn import EpistemicNearestNeighbors


def test_neighbors_returns_correct_number_and_ordering():
    import conftest

    model, train_x, train_y, _, _ = conftest.make_enn_model()
    neighbors = model.neighbors(np.zeros(3, dtype=float), k=5, exclude_nearest=False)
    assert len(neighbors) == 5
    indices = []
    for x_n, y_n in neighbors:
        assert x_n.shape == (3,) and y_n.shape == (1,)
        matches = np.all(np.abs(train_x - x_n[np.newaxis, :]) < 1e-6, axis=1)
        assert np.any(matches), "Neighbor x not found in training data"
        idx = np.where(matches)[0][0]
        indices.append(idx)
        assert np.allclose(train_y[idx], y_n), "Neighbor y doesn't match training data"
    distances = [np.linalg.norm(train_x[i]) for i in indices]
    assert all(
        distances[i] <= distances[i + 1] + 1e-6 for i in range(len(distances) - 1)
    )


def test_neighbors_exclude_nearest():
    import conftest

    model, train_x, _train_y, _train_yvar, _rng = conftest.make_enn_model()
    x_query = train_x[5].copy()
    neighbors_exclude = model.neighbors(x_query, k=5, exclude_nearest=True)
    neighbors_include = model.neighbors(x_query, k=5, exclude_nearest=False)
    assert len(neighbors_exclude) == 5 and len(neighbors_include) == 5
    assert np.allclose(neighbors_include[0][0], x_query)
    assert not np.allclose(neighbors_exclude[0][0], x_query)


def test_neighbors_with_empty_observations():
    d = 3
    train_x = np.zeros((0, d), dtype=float)
    train_y = np.zeros((0, 1), dtype=float)
    train_yvar = np.ones((0, 1), dtype=float)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    x_query = np.zeros(d, dtype=float)
    neighbors = model.neighbors(x_query, k=5, exclude_nearest=False)
    assert neighbors == []


def test_neighbors_k_larger_than_available():
    rng = np.random.default_rng(0)
    n, d = 5, 3
    train_x = rng.standard_normal((n, d))
    train_y = train_x.sum(axis=1, keepdims=True).astype(float)
    train_yvar = 0.1 * np.ones_like(train_y)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    x_query = np.zeros(d, dtype=float)
    neighbors = model.neighbors(x_query, k=20, exclude_nearest=False)
    assert len(neighbors) == n


def test_neighbors_k_zero():
    import conftest

    model, _train_x, _train_y, _train_yvar, _rng = conftest.make_enn_model(n=10)
    x_query = np.zeros(3, dtype=float)
    neighbors = model.neighbors(x_query, k=0, exclude_nearest=False)
    assert neighbors == []


def test_neighbors_with_multiple_metrics():
    rng = np.random.default_rng(0)
    n, d = 15, 3
    train_x = rng.standard_normal((n, d))
    train_y = rng.standard_normal((n, 2))
    train_yvar = 0.1 * np.ones_like(train_y)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    x_query = np.zeros(d, dtype=float)
    neighbors = model.neighbors(x_query, k=5, exclude_nearest=False)
    assert len(neighbors) == 5
    for x_neighbor, y_neighbor in neighbors:
        assert x_neighbor.shape == (d,) and y_neighbor.shape == (2,)


def test_neighbors_accepts_2d_input():
    import conftest

    model, _train_x, _train_y, _train_yvar, _rng = conftest.make_enn_model(n=10)
    d = 3
    x_query_1d = np.zeros(d, dtype=float)
    neighbors_1d = model.neighbors(x_query_1d, k=3, exclude_nearest=False)
    x_query_2d = np.zeros((1, d), dtype=float)
    neighbors_2d = model.neighbors(x_query_2d, k=3, exclude_nearest=False)
    assert len(neighbors_1d) == len(neighbors_2d) == 3
    for (x1, y1), (x2, y2) in zip(neighbors_1d, neighbors_2d):
        assert np.allclose(x1, x2) and np.allclose(y1, y2)


def test_neighbors_exclude_nearest_requires_multiple_observations():
    rng = np.random.default_rng(0)
    d = 3
    train_x = rng.standard_normal((1, d))
    train_y, train_yvar = np.array([[1.0]]), np.array([[0.1]])
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    x_query = np.zeros(d, dtype=float)
    with pytest.raises(
        ValueError, match="exclude_nearest=True requires at least 2 observations"
    ):
        model.neighbors(x_query, k=1, exclude_nearest=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator

from .turbo_config import LHDOnlyConfig


class LHDOnlyImpl:
    def __init__(self, config: LHDOnlyConfig) -> None:
        self._config = config

    @property
    def always_clears_on_restart(self) -> bool:
        return False

    def get_x_center(
        self,
        x_obs_list: list,
        y_obs_list: list,
        rng: Generator,
        tr_state: Any = None,
    ) -> np.ndarray | None:
        return None

    def needs_tr_list(self) -> bool:
        return False

    def try_early_ask(
        self,
        num_arms: int,
        x_obs_list: list,
        draw_initial_fn: Callable[[int], np.ndarray],
        get_init_lhd_points_fn: Callable[[int], np.ndarray],
    ) -> np.ndarray | None:
        return None

    def prepare_ask(
        self,
        x_obs_list: list,
        y_obs_list: list,
        yvar_obs_list: list,
        num_dim: int,
        gp_num_steps: int,
        rng: Any | None = None,
    ) -> tuple[Any, float | None, float | None, np.ndarray | None]:
        return None, None, None, None

    def select_candidates(
        self,
        x_cand: np.ndarray,
        num_arms: int,
        num_dim: int,
        rng: Generator,
        fallback_fn: Callable[[np.ndarray, int], np.ndarray],
        from_unit_fn: Callable[[np.ndarray], np.ndarray],
        tr_state: Any = None,  # noqa: ARG002
    ) -> np.ndarray:
        from .turbo_utils import latin_hypercube

        unit = latin_hypercube(num_arms, num_dim, rng=rng)
        return from_unit_fn(unit)

    def estimate_y(self, x_unit: np.ndarray, y_observed: np.ndarray) -> np.ndarray:
        from .impl_helpers import estimate_y_passthrough

        return estimate_y_passthrough(y_observed)

    def get_mu_sigma(self, x_unit: np.ndarray) -> tuple[np.ndarray, np.ndarray] | None:
        return None



from __future__ import annotations

import argparse
import statistics
import time
from dataclasses import dataclass

import numpy as np


@dataclass(frozen=True)
class BenchResult:
    num_candidates: int
    times_s: list[float]
    error: str | None = None


def _parse_int_list(csv: str) -> list[int]:
    parts = [p.strip() for p in csv.split(",") if p.strip()]
    values = [int(p) for p in parts]
    if not values:
        raise ValueError("no candidates provided")
    if any(v <= 0 for v in values):
        raise ValueError(f"all candidates must be > 0, got {values}")
    return values


def bench_raasp(
    *,
    num_dim: int,
    num_candidates_list: list[int],
    repeats: int,
    seed: int,
) -> list[BenchResult]:
    from enn.turbo.turbo_utils import generate_raasp_candidates
    from scipy.stats import qmc

    if num_dim <= 0:
        raise ValueError(num_dim)
    if repeats <= 0:
        raise ValueError(repeats)

    center = np.full(num_dim, 0.5, dtype=float)
    lb = np.zeros(num_dim, dtype=float)
    ub = np.ones(num_dim, dtype=float)

    results: list[BenchResult] = []
    for num_candidates in num_candidates_list:
        times_s: list[float] = []
        error: str | None = None
        for rep in range(repeats):
            rep_seed = seed + rep
            rng = np.random.default_rng(rep_seed)
            try:
                t0 = time.perf_counter()
                sobol = qmc.Sobol(d=num_dim, scramble=True, seed=rep_seed)
                x = generate_raasp_candidates(
                    center,
                    lb,
                    ub,
                    num_candidates,
                    rng=rng,
                    sobol_engine=sobol,
                )
                # Ensure full materialization (defensive; should already be a numpy array).
                _ = float(np.sum(x))
                times_s.append(time.perf_counter() - t0)
            except MemoryError:
                error = "MemoryError"
                break
        results.append(
            BenchResult(num_candidates=num_candidates, times_s=times_s, error=error)
        )
    return results


def _fmt_stats(times_s: list[float]) -> str:
    if not times_s:
        return "n/a"
    if len(times_s) == 1:
        return f"{times_s[0]:.4f}s"
    return (
        f"min={min(times_s):.4f}s "
        f"median={statistics.median(times_s):.4f}s "
        f"mean={statistics.mean(times_s):.4f}s "
        f"(n={len(times_s)})"
    )


def main() -> None:
    p = argparse.ArgumentParser(
        description="Benchmark RAASP candidate generation time."
    )
    p.add_argument("--num-dim", type=int, default=10_000)
    p.add_argument(
        "--candidates",
        type=str,
        default="50,100,200,500,1000,2000",
        help="Comma-separated list of num_candidates values.",
    )
    p.add_argument("--repeats", type=int, default=3)
    p.add_argument("--seed", type=int, default=0)
    args = p.parse_args()

    num_candidates_list = _parse_int_list(args.candidates)
    results = bench_raasp(
        num_dim=int(args.num_dim),
        num_candidates_list=num_candidates_list,
        repeats=int(args.repeats),
        seed=int(args.seed),
    )

    print(f"num_dim={args.num_dim} repeats={args.repeats} seed={args.seed}")
    print("num_candidates\tseconds")
    for r in results:
        if r.error is not None:
            print(f"{r.num_candidates}\t\tERROR: {r.error}")
        else:
            print(f"{r.num_candidates}\t\t{_fmt_stats(r.times_s)}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator


@dataclass
class TurboTrustRegion:
    num_dim: int
    num_arms: int
    length: float = 0.8
    length_init: float = 0.8
    length_min: float = 0.5**7
    length_max: float = 1.6
    failure_counter: int = 0
    success_counter: int = 0
    best_value: float = -float("inf")
    prev_num_obs: int = 0

    def __post_init__(self) -> None:
        import numpy as np

        self.failure_tolerance = int(
            np.ceil(
                max(
                    4.0 / float(self.num_arms),
                    float(self.num_dim) / float(self.num_arms),
                )
            )
        )
        self.success_tolerance = 3

    def update(self, values: np.ndarray | Any) -> None:
        import numpy as np

        values = np.asarray(values, dtype=float)
        if values.ndim == 2:
            if values.shape[1] != 1:
                raise ValueError(f"TurboTrustRegion expects m=1, got {values.shape}")
            values = values[:, 0]
        elif values.ndim != 1:
            raise ValueError(values.shape)
        if values.size == 0:
            return
        new_values = values[self.prev_num_obs :]
        if new_values.size == 0:
            return
        if not np.isfinite(self.best_value):
            self.best_value = float(np.max(new_values))
            self.prev_num_obs = values.size
            return
        improved = np.max(new_values) > self.best_value + 1e-3 * np.abs(self.best_value)
        if improved:
            self.success_counter += 1
            self.failure_counter = 0
        else:
            self.success_counter = 0
            self.failure_counter += 1
        if self.success_counter >= self.success_tolerance:
            self.length = min(2.0 * self.length, self.length_max)
            self.success_counter = 0
        elif self.failure_counter >= self.failure_tolerance:
            self.length = 0.5 * self.length
            self.failure_counter = 0

        self.best_value = max(self.best_value, float(np.max(new_values)))
        self.prev_num_obs = values.size

    def needs_restart(self) -> bool:
        return self.length < self.length_min

    def restart(self) -> None:
        self.length = self.length_init
        self.failure_counter = 0
        self.success_counter = 0
        self.best_value = -float("inf")
        self.prev_num_obs = 0

    def validate_request(self, num_arms: int, *, is_fallback: bool = False) -> None:
        from .tr_helpers import validate_trust_region_request

        validate_trust_region_request(num_arms, self.num_arms, is_fallback=is_fallback)

    def compute_bounds_1d(
        self, x_center: np.ndarray | Any, lengthscales: np.ndarray | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        import numpy as np

        if lengthscales is None:
            half_length = 0.5 * self.length
        else:
            half_length = lengthscales * self.length / 2.0
        lb = np.clip(x_center - half_length, 0.0, 1.0)
        ub = np.clip(x_center + half_length, 0.0, 1.0)
        return lb, ub

    def get_incumbent_indices(self, y: np.ndarray | Any, rng: Generator) -> np.ndarray:
        import numpy as np

        from .turbo_utils import argmax_random_tie

        y = np.asarray(y, dtype=float)
        y_1d = y[:, 0] if y.ndim == 2 else y
        return np.array([argmax_random_tie(y_1d, rng=rng)])



from __future__ import annotations

import numpy as np

from enn.enn import EpistemicNearestNeighbors
from enn.enn.enn_params import ENNParams


def _params(
    k: int, *, epi_var_scale: float = 1.0, ale_homoscedastic_scale: float = 0.0
):
    return ENNParams(
        k=int(k),
        epi_var_scale=float(epi_var_scale),
        ale_homoscedastic_scale=float(ale_homoscedastic_scale),
    )


def _make_single_metric_train_data(*, rng, n: int, d: int, noise_std: float):
    train_x = rng.standard_normal((n, d))
    train_y = train_x.sum(axis=1, keepdims=True) + rng.standard_normal((n, 1)) * float(
        noise_std
    )
    return train_x, train_y, 0.1 * np.ones_like(train_y)


def test_epistemic_nearest_neighbors_scale_invariance():
    rng = np.random.default_rng(42)
    train_x, train_y, train_yvar = _make_single_metric_train_data(
        rng=rng, n=20, d=3, noise_std=0.1
    )
    model_base = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    model_scaled = EpistemicNearestNeighbors(
        train_x, train_y * 100.0, train_yvar * 10000.0
    )
    x_test, params = rng.standard_normal((10, 3)), _params(5)
    post_base = model_base.posterior(x_test, params=params)
    post_scaled = model_scaled.posterior(x_test, params=params)
    assert np.allclose(post_scaled.mu, post_base.mu * 100.0, rtol=1e-10)
    assert np.allclose(post_scaled.se, post_base.se * 100.0, rtol=1e-10)


def test_epistemic_nearest_neighbors_shift_invariance():
    rng = np.random.default_rng(42)
    train_x, train_y, train_yvar = _make_single_metric_train_data(
        rng=rng, n=20, d=3, noise_std=0.1
    )
    model_base = EpistemicNearestNeighbors(train_x, train_y, train_yvar)
    model_shifted = EpistemicNearestNeighbors(train_x, train_y + 1000.0, train_yvar)
    x_test, params = rng.standard_normal((10, 3)), _params(5)
    post_base = model_base.posterior(x_test, params=params)
    post_shifted = model_shifted.posterior(x_test, params=params)
    assert np.allclose(post_shifted.mu, post_base.mu + 1000.0, rtol=1e-10)
    assert np.allclose(post_shifted.se, post_base.se, rtol=1e-10)


def test_epistemic_nearest_neighbors_x_rescaling_is_invariant_when_scale_x_enabled():
    rng = np.random.default_rng(0)
    train_x = rng.standard_normal((50, 4))
    train_y = train_x.sum(axis=1, keepdims=True)
    scale = np.array([[100.0, 0.1, 3.0, 1.0]])
    x_test = rng.standard_normal((10, 4))
    params = ENNParams(k=7, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    model = EpistemicNearestNeighbors(
        train_x, train_y, 0.1 * np.ones_like(train_y), scale_x=True
    )
    model_scaled = EpistemicNearestNeighbors(
        train_x * scale, train_y, 0.1 * np.ones_like(train_y), scale_x=True
    )
    post = model.posterior(x_test, params=params)
    post_scaled = model_scaled.posterior(x_test * scale, params=params)
    assert np.allclose(post.mu, post_scaled.mu, rtol=1e-6, atol=1e-8)
    assert np.allclose(post.se, post_scaled.se, rtol=1e-6, atol=1e-8)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np


@dataclass
class NoTrustRegion:
    num_dim: int
    num_arms: int
    length: float = 1.0

    def update(self, values: np.ndarray | Any) -> None:
        return

    def needs_restart(self) -> bool:
        return False

    def restart(self) -> None:
        return

    def validate_request(self, num_arms: int, *, is_fallback: bool = False) -> None:
        from .tr_helpers import validate_trust_region_request

        validate_trust_region_request(num_arms, self.num_arms, is_fallback=is_fallback)

    def compute_bounds_1d(
        self, x_center: np.ndarray | Any, lengthscales: np.ndarray | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        import numpy as np

        lb = np.zeros_like(x_center, dtype=float)
        ub = np.ones_like(x_center, dtype=float)
        return lb, ub



from __future__ import annotations

import contextlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Iterator

import numpy as np

if TYPE_CHECKING:
    import torch
    from numpy.random import Generator
    from scipy.stats._qmc import QMCEngine


__all__ = [
    "Telemetry",
]


@dataclass(frozen=True)
class Telemetry:
    dt_fit: float
    dt_sel: float
    dt_gen: float = 0.0
    dt_tell: float = 0.0


@contextlib.contextmanager
def record_duration(set_dt: Callable[[float], None]) -> Iterator[None]:
    import time

    t0 = time.perf_counter()
    try:
        yield
    finally:
        set_dt(time.perf_counter() - t0)


def _next_power_of_2(n: int) -> int:
    if n <= 0:
        return 1
    return 1 << (n - 1).bit_length()


@contextlib.contextmanager
def torch_seed_context(
    seed: int, device: torch.device | Any | None = None
) -> Iterator[None]:
    import torch

    devices: list[int] | None = None
    if device is not None and getattr(device, "type", None) == "cuda":
        idx = 0 if getattr(device, "index", None) is None else int(device.index)
        devices = [idx]
    with torch.random.fork_rng(devices=devices, enabled=True):
        torch.manual_seed(int(seed))
        if device is not None and getattr(device, "type", None) == "cuda":
            torch.cuda.manual_seed_all(int(seed))
        if device is not None and getattr(device, "type", None) == "mps":
            if hasattr(torch, "mps") and hasattr(torch.mps, "manual_seed"):
                torch.mps.manual_seed(int(seed))
        yield


def get_gp_posterior_suppress_warning(model: Any, x_torch: Any) -> Any:
    import warnings

    try:
        from gpytorch.utils.warnings import GPInputWarning
    except Exception:  # pragma: no cover
        GPInputWarning = None

    if GPInputWarning is None:
        return model.posterior(x_torch)
    # We intentionally evaluate the GP posterior at the training inputs
    # (observed points) when choosing the center. GPyTorch warns about this
    # in debug mode, but it's expected for our usage.
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"The input matches the stored training data\..*",
            category=GPInputWarning,
        )
        return model.posterior(x_torch)


def latin_hypercube(
    num_points: int, num_dim: int, *, rng: Generator | Any
) -> np.ndarray:
    x = np.zeros((num_points, num_dim))
    centers = (1.0 + 2.0 * np.arange(0.0, num_points)) / float(2 * num_points)
    for j in range(num_dim):
        x[:, j] = centers[rng.permutation(num_points)]
    pert = rng.uniform(-1.0, 1.0, size=(num_points, num_dim)) / float(2 * num_points)
    x += pert
    return x


def argmax_random_tie(values: np.ndarray | Any, *, rng: Generator | Any) -> int:
    if values.ndim != 1:
        raise ValueError(values.shape)
    max_val = float(np.max(values))
    idx = np.nonzero(values >= max_val)[0]
    if idx.size == 0:
        return int(rng.integers(values.size))
    if idx.size == 1:
        return int(idx[0])
    j = int(rng.integers(idx.size))
    return int(idx[j])


def sobol_perturb_np(
    x_center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    mask: np.ndarray | Any,
    *,
    sobol_engine: QMCEngine | Any,
) -> np.ndarray:
    n_sobol = _next_power_of_2(num_candidates)
    sobol_samples = sobol_engine.random(n_sobol)[:num_candidates]
    lb_array = np.asarray(lb)
    ub_array = np.asarray(ub)
    pert = lb_array + (ub_array - lb_array) * sobol_samples
    candidates = np.tile(x_center, (num_candidates, 1))
    if np.any(mask):
        candidates[mask] = pert[mask]
    return candidates


def uniform_perturb_np(
    x_center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    mask: np.ndarray | Any,
    *,
    rng: Generator | Any,
) -> np.ndarray:
    lb_array = np.asarray(lb)
    ub_array = np.asarray(ub)
    pert = lb_array + (ub_array - lb_array) * rng.uniform(
        0.0, 1.0, size=(num_candidates, x_center.shape[-1])
    )
    candidates = np.tile(x_center, (num_candidates, 1))
    if np.any(mask):
        candidates[mask] = pert[mask]
    return candidates


def _raasp_mask(
    *,
    num_candidates: int,
    num_dim: int,
    num_pert: int,
    rng: Generator | Any,
) -> np.ndarray:
    prob_perturb = min(num_pert / num_dim, 1.0)
    mask = rng.random((num_candidates, num_dim)) <= prob_perturb
    ind = np.nonzero(~mask.any(axis=1))[0]
    if len(ind) > 0:
        mask[ind, rng.integers(0, num_dim, size=len(ind))] = True
    return mask


def raasp(
    x_center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    *,
    num_pert: int = 20,
    rng: Generator | Any,
    sobol_engine: QMCEngine | Any,
) -> np.ndarray:
    num_dim = x_center.shape[-1]
    mask = _raasp_mask(
        num_candidates=num_candidates, num_dim=num_dim, num_pert=num_pert, rng=rng
    )
    return sobol_perturb_np(
        x_center, lb, ub, num_candidates, mask, sobol_engine=sobol_engine
    )


def raasp_uniform(
    x_center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    *,
    num_pert: int = 20,
    rng: Generator | Any,
) -> np.ndarray:
    num_dim = x_center.shape[-1]
    mask = _raasp_mask(
        num_candidates=num_candidates, num_dim=num_dim, num_pert=num_pert, rng=rng
    )
    return uniform_perturb_np(x_center, lb, ub, num_candidates, mask, rng=rng)


def generate_raasp_candidates(
    center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    *,
    rng: Generator | Any,
    sobol_engine: QMCEngine | Any,
    num_pert: int = 20,
) -> np.ndarray:
    if num_candidates <= 0:
        raise ValueError(num_candidates)
    return raasp(
        center,
        lb,
        ub,
        num_candidates,
        num_pert=num_pert,
        rng=rng,
        sobol_engine=sobol_engine,
    )


def generate_raasp_candidates_uniform(
    center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    *,
    rng: Generator | Any,
    num_pert: int = 20,
) -> np.ndarray:
    if num_candidates <= 0:
        raise ValueError(num_candidates)
    return raasp_uniform(center, lb, ub, num_candidates, num_pert=num_pert, rng=rng)


def to_unit(x: np.ndarray | Any, bounds: np.ndarray | Any) -> np.ndarray:
    lb = bounds[:, 0]
    ub = bounds[:, 1]
    if np.any(ub <= lb):
        raise ValueError(bounds)
    return (x - lb) / (ub - lb)


def from_unit(x_unit: np.ndarray | Any, bounds: np.ndarray | Any) -> np.ndarray:
    lb = np.asarray(bounds[:, 0])
    ub = np.asarray(bounds[:, 1])
    return lb + x_unit * (ub - lb)


def gp_thompson_sample(
    model: Any,
    x_cand: np.ndarray | Any,
    num_arms: int,
    rng: Generator | Any,
    *,
    gp_y_mean: float,
    gp_y_std: float,
) -> np.ndarray:
    import gpytorch
    import torch

    x_torch = torch.as_tensor(x_cand, dtype=torch.float64)
    seed = int(rng.integers(2**31 - 1))
    with (
        torch.no_grad(),
        gpytorch.settings.fast_pred_var(),
        torch_seed_context(seed, device=x_torch.device),
    ):
        posterior = model.posterior(x_torch)
        samples = posterior.sample(sample_shape=torch.Size([1]))
    if samples.ndim != 2:
        raise ValueError(samples.shape)
    ts = samples[0].reshape(-1)
    scores = ts.detach().cpu().numpy().reshape(-1)
    scores = gp_y_mean + gp_y_std * scores
    shuffled_indices = rng.permutation(len(scores))
    shuffled_scores = scores[shuffled_indices]
    top_k_in_shuffled = np.argpartition(-shuffled_scores, num_arms - 1)[:num_arms]
    idx = shuffled_indices[top_k_in_shuffled]
    return idx



from .turbo_mode import TurboMode
from .turbo_optimizer import TurboOptimizer
from .turbo_utils import Telemetry

Turbo = TurboOptimizer

__all__ = [
    "TurboMode",
    "TurboOptimizer",
    "Turbo",
    "Telemetry",
]



from __future__ import annotations

import numpy as np
import pytest

from enn.turbo.turbo_config import (
    LHDOnlyConfig,
    TurboConfig,
    TurboENNConfig,
    TurboOneConfig,
)
from enn.turbo.turbo_enn_impl import TurboENNImpl
from enn.turbo.turbo_mode import TurboMode
from enn.turbo.turbo_one_impl import TurboOneImpl
from enn.turbo.turbo_optimizer import TurboOptimizer


def test_lhd_only_runs_after_initial_observations():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    config = LHDOnlyConfig(num_init=4)
    opt = TurboOptimizer(bounds=bounds, mode=TurboMode.LHD_ONLY, rng=rng, config=config)
    for _ in range(3):
        x = opt.ask(num_arms=2)
        y = -np.sum(x**2, axis=1)
        opt.tell(x, y)
    x_after_init = opt.ask(num_arms=2)
    assert x_after_init.shape == (2, 2)


def test_turbo_enn_impl_get_x_center():
    rng = np.random.default_rng(42)
    config = TurboENNConfig(k=5)
    impl = TurboENNImpl(config)
    assert impl.get_x_center([], [], rng) is None

    x_obs = rng.random((20, 3)).tolist()
    y_obs = [float(i) for i in range(20)]
    result_before_fit = impl.get_x_center(x_obs, y_obs, rng)
    assert result_before_fit is not None
    assert np.allclose(result_before_fit, np.asarray(x_obs, dtype=float)[19])

    impl.prepare_ask(x_obs, y_obs, [0.0] * 20, num_dim=3, gp_num_steps=10, rng=rng)
    result_after_fit = impl.get_x_center(x_obs, y_obs, rng)
    assert result_after_fit is not None and result_after_fit.shape == (3,)
    x_array = np.asarray(x_obs, dtype=float)
    top_5_indices = np.argsort(y_obs)[-5:]
    assert any(
        np.allclose(result_after_fit, x_array[top_5_indices[i]]) for i in range(5)
    )


def test_turbo_one_impl_get_x_center_requires_fit_for_n_ge_2():
    rng = np.random.default_rng(0)
    impl = TurboOneImpl(TurboOneConfig())
    x1 = rng.random((1, 3)).tolist()
    center1 = impl.get_x_center(x1, [1.0], rng)
    assert center1 is not None and np.allclose(center1, np.asarray(x1, dtype=float)[0])

    x2 = rng.random((2, 3)).tolist()
    with pytest.raises(RuntimeError, match="prepare_ask"):
        impl.get_x_center(x2, [0.0, 1.0], rng)

    impl.prepare_ask(x2, [0.0, 1.0], [], num_dim=3, gp_num_steps=5, rng=rng)
    center2 = impl.get_x_center(x2, [0.0, 1.0], rng)
    assert center2 is not None
    x2_array = np.asarray(x2, dtype=float)
    assert np.allclose(center2, x2_array[0]) or np.allclose(center2, x2_array[1])


def test_turbo_config_num_metrics_validation():
    TurboConfig(tr_type="turbo", num_metrics=None)
    TurboConfig(tr_type="turbo", num_metrics=1)
    with pytest.raises(ValueError, match="num_metrics must be 1 for tr_type='turbo'"):
        TurboConfig(tr_type="turbo", num_metrics=2)
    TurboConfig(tr_type="none", num_metrics=None)
    TurboConfig(tr_type="none", num_metrics=1)
    with pytest.raises(ValueError, match="num_metrics must be 1 for tr_type='none'"):
        TurboConfig(tr_type="none", num_metrics=2)
    TurboConfig(tr_type="morbo", num_metrics=None)
    TurboConfig(tr_type="morbo", num_metrics=2)
    with pytest.raises(ValueError, match="num_metrics must be >= 1"):
        TurboConfig(tr_type="morbo", num_metrics=0)



from __future__ import annotations

import numpy as np
import pytest
from scipy.stats import qmc

from enn.turbo.morbo_trust_region import MorboTrustRegion
from enn.turbo.no_trust_region import NoTrustRegion
from enn.turbo.tr_helpers import validate_trust_region_request
from enn.turbo.turbo_trust_region import TurboTrustRegion


def test_no_trust_region_init():
    tr = NoTrustRegion(num_dim=3, num_arms=4)
    assert tr.num_dim == 3 and tr.num_arms == 4 and tr.length == 1.0


def test_no_trust_region_update_does_nothing():
    tr = NoTrustRegion(num_dim=3, num_arms=4)
    tr.update(np.array([1.0, 2.0, 3.0]))
    assert tr.length == 1.0


def test_no_trust_region_needs_restart():
    tr = NoTrustRegion(num_dim=3, num_arms=4)
    assert not tr.needs_restart()


def test_no_trust_region_validate_request():
    tr = NoTrustRegion(num_dim=3, num_arms=4)
    tr.validate_request(4)
    with pytest.raises(ValueError):
        tr.validate_request(5)


def test_no_trust_region_compute_bounds_1d():
    tr = NoTrustRegion(num_dim=3, num_arms=4)
    x_center = np.array([0.5, 0.5, 0.5])
    lb, ub = tr.compute_bounds_1d(x_center)
    assert np.allclose(lb, 0.0) and np.allclose(ub, 1.0)


def test_no_trust_region_generate_candidates():
    from enn.turbo.tr_helpers import generate_tr_candidates

    tr = NoTrustRegion(num_dim=3, num_arms=4)
    rng = np.random.default_rng(42)
    sobol = qmc.Sobol(d=3, scramble=True, seed=42)
    x_center = np.array([0.5, 0.5, 0.5])
    candidates = generate_tr_candidates(
        tr.compute_bounds_1d, x_center, None, 100, rng=rng, sobol_engine=sobol
    )
    assert candidates.shape == (100, 3)
    assert np.all(candidates >= 0.0) and np.all(candidates <= 1.0)


def test_turbo_trust_region_validate_request():
    tr = TurboTrustRegion(num_dim=3, num_arms=4)
    tr.validate_request(4)
    with pytest.raises(ValueError):
        tr.validate_request(5)


def test_turbo_trust_region_get_incumbent_indices():
    tr = TurboTrustRegion(num_dim=3, num_arms=4)
    rng = np.random.default_rng(42)
    y = np.array([1.0, 5.0, 3.0, 2.0, 4.0])
    indices = tr.get_incumbent_indices(y, rng)
    assert 1 in indices


def test_morbo_trust_region_validate_request():
    rng = np.random.default_rng(42)
    tr = MorboTrustRegion(num_dim=3, num_arms=4, num_metrics=2, rng=rng)
    tr.validate_request(4)
    with pytest.raises(ValueError):
        tr.validate_request(5)


def test_morbo_trust_region_get_incumbent_indices():
    rng = np.random.default_rng(42)
    tr = MorboTrustRegion(num_dim=3, num_arms=4, num_metrics=2, rng=rng)
    y = np.array([[1.0, 5.0], [5.0, 1.0], [3.0, 3.0], [2.0, 2.0]])
    indices = tr.get_incumbent_indices(y, rng)
    assert len(indices) >= 1


def test_validate_trust_region_request_exact():
    validate_trust_region_request(4, 4)
    with pytest.raises(ValueError):
        validate_trust_region_request(3, 4)


def test_validate_trust_region_request_fallback():
    validate_trust_region_request(3, 4, is_fallback=True)
    with pytest.raises(ValueError):
        validate_trust_region_request(5, 4, is_fallback=True)



from __future__ import annotations

import numpy as np
from scipy.stats import qmc

from enn.turbo.morbo_trust_region import MorboTrustRegion
from enn.turbo.turbo_trust_region import TurboTrustRegion


def test_trust_region_state_update_and_restart_and_bounds():
    state = TurboTrustRegion(num_dim=2, num_arms=2)
    values = []
    for v in [0.0, 1.0, 2.0]:
        values.append(v)
        state.update(np.array(values, dtype=float))
    x_center = np.zeros(2, dtype=float)
    lb, ub = state.compute_bounds_1d(x_center)
    assert lb.shape == (2,) and ub.shape == (2,)
    state.length = state.length_min / 2.0
    assert state.needs_restart()
    state.restart()
    assert state.length == state.length_init


def test_morbo_chebyshev_trust_region_weights_and_scaling():
    rng1, rng2 = np.random.default_rng(0), np.random.default_rng(0)
    tr1 = MorboTrustRegion(num_dim=3, num_arms=1, num_metrics=2, rng=rng1)
    tr2 = MorboTrustRegion(num_dim=3, num_arms=1, num_metrics=2, rng=rng2)
    assert np.allclose(tr1.weights, tr2.weights)
    assert tr1.weights.shape == (2,) and np.all(tr1.weights > 0.0)
    assert np.isclose(tr1.weights.sum(), 1.0)

    y_obs = np.array([[1.0, 1.0], [1.0, 1.0]], dtype=float)
    tr1.update(y_obs)
    scores = tr1.scalarize(y_obs, clip=True)
    assert scores.shape == (2,) and np.all(np.isfinite(scores))
    t = 0.5 * tr1.weights.reshape(1, -1)
    expected = np.min(t, axis=1) + 0.05 * np.sum(t, axis=1)
    assert np.allclose(scores[:1], expected)

    sobol_engine = qmc.Sobol(d=3, scramble=True, seed=0)
    x_center = np.array([0.5, 0.5, 0.5], dtype=float)
    x_cand = tr1.generate_candidates(
        x_center=x_center,
        lengthscales=None,
        num_candidates=64,
        rng=np.random.default_rng(1),
        sobol_engine=sobol_engine,
    )
    assert x_cand.shape == (64, 3) and np.all(x_cand >= 0.0) and np.all(x_cand <= 1.0)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator

    from .enn import EpistemicNearestNeighbors
    from .enn_params import ENNParams


def _validate_subsample_inputs(
    x: np.ndarray | Any, y: np.ndarray | Any, P: int, paramss: list
) -> tuple[np.ndarray, np.ndarray]:
    import numpy as np

    x_array = np.asarray(x, dtype=float)
    if x_array.ndim != 2:
        raise ValueError(x_array.shape)
    y_array = np.asarray(y, dtype=float)
    if y_array.ndim == 1:
        y_array = y_array.reshape(-1, 1)
    if y_array.ndim != 2:
        raise ValueError(y_array.shape)
    if x_array.shape[0] != y_array.shape[0]:
        raise ValueError((x_array.shape, y_array.shape))
    if P <= 0:
        raise ValueError(P)
    if len(paramss) == 0:
        raise ValueError("paramss must be non-empty")
    return x_array, y_array


def _compute_single_loglik(
    y_scaled: np.ndarray, mu_i: np.ndarray, se_i: np.ndarray
) -> float:
    import numpy as np

    if not np.isfinite(mu_i).all() or not np.isfinite(se_i).all():
        return 0.0
    if np.any(se_i <= 0.0):
        return 0.0
    var_scaled = se_i**2
    loglik = -0.5 * np.sum(
        np.log(2.0 * np.pi * var_scaled) + (y_scaled - mu_i) ** 2 / var_scaled
    )
    return float(loglik) if np.isfinite(loglik) else 0.0


def subsample_loglik(
    model: EpistemicNearestNeighbors | Any,
    x: np.ndarray | Any,
    y: np.ndarray | Any,
    *,
    paramss: list[ENNParams] | list[Any],
    P: int = 10,
    rng: Generator | Any,
) -> list[float]:
    import numpy as np

    x_array, y_array = _validate_subsample_inputs(x, y, P, paramss)
    n = x_array.shape[0]
    if n == 0 or len(model) <= 1:
        return [0.0] * len(paramss)
    P_actual = min(P, n)
    indices = (
        np.arange(n, dtype=int) if P_actual == n else rng.permutation(n)[:P_actual]
    )
    x_sel, y_sel = x_array[indices], y_array[indices]
    if not np.isfinite(y_sel).all():
        return [0.0] * len(paramss)

    from .enn_params import PosteriorFlags

    post = model.batch_posterior(
        x_sel,
        paramss,
        flags=PosteriorFlags(exclude_nearest=True, observation_noise=True),
    )
    num_params, num_outputs = len(paramss), y_sel.shape[1]
    expected_shape = (num_params, P_actual, num_outputs)
    if post.mu.shape != expected_shape or post.se.shape != expected_shape:
        raise ValueError((post.mu.shape, post.se.shape, expected_shape))

    y_std = np.std(y_array, axis=0, keepdims=True).astype(float)
    y_std = np.where(np.isfinite(y_std) & (y_std > 0.0), y_std, 1.0)
    y_scaled, mu_scaled, se_scaled = y_sel / y_std, post.mu / y_std, post.se / y_std
    return [
        _compute_single_loglik(y_scaled, mu_scaled[i], se_scaled[i])
        for i in range(num_params)
    ]


def enn_fit(
    model: EpistemicNearestNeighbors | Any,
    *,
    k: int,
    num_fit_candidates: int,
    num_fit_samples: int = 10,
    rng: Generator | Any,
    params_warm_start: ENNParams | Any | None = None,
) -> ENNParams:
    from .enn_params import ENNParams

    train_x = model.train_x
    train_y = model.train_y
    log_min = -3.0
    log_max = 3.0
    epi_var_scale_log_values = rng.uniform(log_min, log_max, size=num_fit_candidates)
    epi_var_scale_values = 10**epi_var_scale_log_values
    ale_homoscedastic_log_values = rng.uniform(
        log_min, log_max, size=num_fit_candidates
    )
    ale_homoscedastic_values = 10**ale_homoscedastic_log_values
    paramss = [
        ENNParams(
            k=k,
            epi_var_scale=float(epi_val),
            ale_homoscedastic_scale=float(ale_val),
        )
        for epi_val, ale_val in zip(epi_var_scale_values, ale_homoscedastic_values)
    ]
    if params_warm_start is not None:
        paramss.append(
            ENNParams(
                k=k,
                epi_var_scale=params_warm_start.epi_var_scale,
                ale_homoscedastic_scale=params_warm_start.ale_homoscedastic_scale,
            )
        )
    if len(paramss) == 0:
        return ENNParams(k=k, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    import numpy as np

    logliks = subsample_loglik(
        model, train_x, train_y, paramss=paramss, P=num_fit_samples, rng=rng
    )
    if len(logliks) == 0:
        return paramss[0]
    best_idx = int(np.argmax(logliks))
    return paramss[best_idx]



from __future__ import annotations

import numpy as np

from enn.enn import EpistemicNearestNeighbors
from enn.enn.enn_params import ENNParams, PosteriorFlags


def test_conditional_posterior_matches_posterior_when_no_whatifs():
    import conftest

    model, _train_x, _train_y, _train_yvar, rng = conftest.make_enn_model()
    x_test = rng.standard_normal((6, 3))
    params = ENNParams(k=5, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    flags = PosteriorFlags(exclude_nearest=True, observation_noise=True)

    x_whatif = np.zeros((0, 3), dtype=float)
    y_whatif = np.zeros((0, 1), dtype=float)

    post = model.posterior(x_test, params=params, flags=flags)
    post_cond = model.conditional_posterior(
        x_whatif, y_whatif, x_test, params=params, flags=flags
    )
    assert np.allclose(post.mu, post_cond.mu)
    assert np.allclose(post.se, post_cond.se)


def test_conditional_posterior_includes_whatif_points_and_does_not_mutate_index():
    d = 3
    train_x = np.array([[10.0, 0.0, 0.0], [0.0, 10.0, 0.0]], dtype=float)
    train_y = np.array([[0.0], [0.0]], dtype=float)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar=None)

    x_test = np.zeros((1, d), dtype=float)
    x_whatif = np.zeros((1, d), dtype=float)
    y_whatif = np.array([[5.0]], dtype=float)

    neighbors_before = model.neighbors(x_test, k=2, exclude_nearest=False)
    post_cond = model.conditional_posterior(
        x_whatif,
        y_whatif,
        x_test,
        params=ENNParams(k=1, epi_var_scale=1.0, ale_homoscedastic_scale=0.0),
        flags=PosteriorFlags(exclude_nearest=False, observation_noise=False),
    )
    neighbors_after = model.neighbors(x_test, k=2, exclude_nearest=False)

    assert len(neighbors_before) == len(neighbors_after)
    for (xb, yb), (xa, ya) in zip(neighbors_before, neighbors_after):
        assert np.allclose(xb, xa)
        assert np.allclose(yb, ya)
    assert np.allclose(post_cond.mu, 5.0)


def test_conditional_posterior_exclude_nearest_drops_whatif_point():
    d = 3
    train_x = np.array([[10.0, 0.0, 0.0], [0.0, 10.0, 0.0]], dtype=float)
    train_y = np.array([[1.0], [2.0]], dtype=float)
    train_yvar = 0.1 * np.ones_like(train_y)
    model = EpistemicNearestNeighbors(train_x, train_y, train_yvar=train_yvar)

    x_test = np.zeros((1, d), dtype=float)
    x_whatif = np.zeros((1, d), dtype=float)
    y_whatif = np.array([[100.0]], dtype=float)
    params = ENNParams(k=1, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)

    post_incl = model.conditional_posterior(
        x_whatif,
        y_whatif,
        x_test,
        params=params,
        flags=PosteriorFlags(exclude_nearest=False, observation_noise=True),
    )
    post_excl = model.conditional_posterior(
        x_whatif,
        y_whatif,
        x_test,
        params=params,
        flags=PosteriorFlags(exclude_nearest=True, observation_noise=True),
    )

    assert np.allclose(post_incl.mu, 100.0)
    assert not np.allclose(post_excl.mu, 100.0)


def test_conditional_posterior_matches_augmented_model_exactly():
    rng = np.random.default_rng(0)
    n_train, n_whatif, d = 8, 3, 3
    train_x = rng.standard_normal((n_train, d))
    train_y = train_x.sum(axis=1, keepdims=True).astype(float)

    x_whatif = rng.standard_normal((n_whatif, d))
    y_whatif = x_whatif.sum(axis=1, keepdims=True).astype(float)

    x_test = rng.standard_normal((5, d))
    params = ENNParams(k=4, epi_var_scale=1.0, ale_homoscedastic_scale=0.0)
    flags = PosteriorFlags(exclude_nearest=True, observation_noise=True)

    enn_a = EpistemicNearestNeighbors(train_x, train_y, train_yvar=None)
    post_a = enn_a.conditional_posterior(
        x_whatif, y_whatif, x_test, params=params, flags=flags
    )

    enn_b = EpistemicNearestNeighbors(
        np.concatenate([train_x, x_whatif], axis=0),
        np.concatenate([train_y, y_whatif], axis=0),
        train_yvar=None,
    )
    post_b = enn_b.posterior(x_test, params=params, flags=flags)

    assert np.allclose(post_a.mu, post_b.mu)
    assert np.allclose(post_a.se, post_b.se)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np


def normal_hash_batch_multi_seed(
    function_seeds: np.ndarray, data_indices: np.ndarray, num_metrics: int
) -> np.ndarray:
    import numpy as np
    from scipy.special import ndtri

    num_seeds = len(function_seeds)
    unique_indices = np.unique(data_indices)
    num_unique = len(unique_indices)
    max_idx = int(unique_indices.max()) + 1

    # Build grids for (seed, unique_idx, metric) combinations
    seed_grid, idx_grid, metric_grid = np.meshgrid(
        function_seeds.astype(np.uint64),
        unique_indices.astype(np.uint64),
        np.arange(num_metrics, dtype=np.uint64),
        indexing="ij",
    )
    seed_flat = seed_grid.ravel()
    idx_flat = idx_grid.ravel()
    metric_flat = metric_grid.ravel()

    combined_seeds = (seed_flat * np.uint64(1_000_003) + idx_flat) * np.uint64(
        1_000_003
    ) + metric_flat

    # Generate uniform values
    uniform_vals = np.empty(len(combined_seeds), dtype=float)
    for i, seed in enumerate(combined_seeds):
        rng = np.random.Generator(np.random.Philox(int(seed)))
        uniform_vals[i] = rng.random()
    uniform_vals = np.clip(uniform_vals, 1e-10, 1.0 - 1e-10)
    normal_vals = ndtri(uniform_vals).reshape(num_seeds, num_unique, num_metrics)

    # Build lookup table per seed and use vectorized indexing
    lookup = np.zeros((num_seeds, max_idx, num_metrics), dtype=float)
    lookup[:, unique_indices, :] = normal_vals

    return lookup[:, data_indices, :]



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal


@dataclass(frozen=True)
class TurboConfig:
    k: int | None = None
    num_candidates: int | None = None
    num_init: int | None = None

    # Experimental
    trailing_obs: int | None = None
    tr_type: Literal["turbo", "morbo", "none"] = "turbo"
    num_metrics: int | None = None
    candidate_rv: Literal["sobol", "uniform"] = "sobol"

    def __post_init__(self) -> None:
        if self.tr_type not in ["turbo", "morbo", "none"]:
            raise ValueError(
                f"tr_type must be 'turbo', 'morbo', or 'none', got {self.tr_type!r}"
            )
        if self.candidate_rv not in ["sobol", "uniform"]:
            raise ValueError(
                f"candidate_rv must be 'sobol' or 'uniform', got {self.candidate_rv!r}"
            )
        if self.num_metrics is not None and self.num_metrics < 1:
            raise ValueError(f"num_metrics must be >= 1, got {self.num_metrics}")
        if self.tr_type == "turbo":
            if self.num_metrics is not None and self.num_metrics != 1:
                raise ValueError(
                    f"num_metrics must be 1 for tr_type='turbo', got {self.num_metrics}"
                )
        if self.tr_type == "none":
            if self.num_metrics is not None and self.num_metrics != 1:
                raise ValueError(
                    f"num_metrics must be 1 for tr_type='none', got {self.num_metrics}"
                )


@dataclass(frozen=True)
class TurboOneConfig(TurboConfig):
    pass


@dataclass(frozen=True)
class TurboZeroConfig(TurboConfig):
    pass


@dataclass(frozen=True)
class LHDOnlyConfig(TurboConfig):
    pass


@dataclass(frozen=True)
class TurboENNConfig(TurboConfig):
    acq_type: Literal["thompson", "pareto", "ucb"] = "pareto"
    num_fit_samples: int | None = None
    num_fit_candidates: int | None = None
    scale_x: bool = False

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.acq_type not in ["thompson", "pareto", "ucb"]:
            raise ValueError(
                f"acq_type must be 'thompson', 'pareto', or 'ucb', got {self.acq_type!r}"
            )
        if self.num_fit_samples is None and self.acq_type != "pareto":
            raise ValueError(f"num_fit_samples required for acq_type={self.acq_type!r}")
        if self.num_fit_samples is not None and int(self.num_fit_samples) <= 0:
            raise ValueError(f"num_fit_samples must be > 0, got {self.num_fit_samples}")
        if self.num_fit_candidates is not None and int(self.num_fit_candidates) <= 0:
            raise ValueError(
                f"num_fit_candidates must be > 0, got {self.num_fit_candidates}"
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np

from . import turbo_optimizer_utils, turbo_utils
from .turbo_mode_registry import make_impl, validate_config

if TYPE_CHECKING:
    from typing import Callable
    from numpy.random import Generator
    from .turbo_config import TurboConfig
    from .turbo_mode import TurboMode


class TurboOptimizer:
    def _init_obs_lists(self) -> None:
        self._x_obs_list: list[list[float]] = []
        self._y_obs_list: list[float] | list[list[float]] = []
        self._y_tr_list: list[float] | list[list[float]] = []
        self._yvar_obs_list: list[float] | list[list[float]] = []
        self._expects_yvar: bool | None = None

    def _validate_init_params(self) -> None:
        if self._num_candidates <= 0:
            raise ValueError(self._num_candidates)
        if self._k is not None and self._k < 3:
            raise ValueError(f"k must be >= 3, got {self._k}")
        if self._trailing_obs is not None and self._trailing_obs <= 0:
            raise ValueError(f"trailing_obs must be > 0, got {self._trailing_obs}")
        if self._num_init <= 0:
            raise ValueError(f"num_init must be > 0, got {self._num_init}")

    def __init__(
        self,
        bounds: np.ndarray,
        mode: TurboMode,
        *,
        rng: Generator,
        config: TurboConfig | None = None,
    ) -> None:
        config = validate_config(mode, config)
        self._config = config
        bounds = np.asarray(bounds, dtype=float)
        if bounds.ndim != 2 or bounds.shape[1] != 2:
            raise ValueError(bounds.shape)
        self._bounds, self._num_dim, self._mode, self._rng = (
            bounds,
            bounds.shape[0],
            mode,
            rng,
        )
        self._num_candidates = int(
            config.num_candidates or min(5000, 100 * self._num_dim)
        )
        self._sobol_seed_base = int(rng.integers(2**31 - 1))
        self._init_obs_lists()
        self._mode_impl: Any = make_impl(mode, config)
        self._tr_state: Any | None = None
        self._gp_num_steps: int = 50
        self._k = None if config.k is None else int(config.k)
        self._trailing_obs = (
            None if config.trailing_obs is None else int(config.trailing_obs)
        )
        self._num_init = int(
            config.num_init if config.num_init is not None else 2 * self._num_dim
        )
        self._validate_init_params()
        self._init_lhd = turbo_utils.from_unit(
            turbo_utils.latin_hypercube(self._num_init, self._num_dim, rng=rng),
            self._bounds,
        )
        self._init_idx, self._dt_fit, self._dt_gen, self._dt_sel, self._dt_tell = (
            0,
            0.0,
            0.0,
            0.0,
            0.0,
        )

    @property
    def tr_obs_count(self) -> int:
        return len(self._y_obs_list)

    @property
    def tr_length(self) -> float | None:
        return (
            None
            if self._tr_state is None or not hasattr(self._tr_state, "length")
            else float(self._tr_state.length)
        )

    def telemetry(self) -> turbo_utils.Telemetry:
        return turbo_utils.Telemetry(
            dt_fit=self._dt_fit,
            dt_gen=self._dt_gen,
            dt_sel=self._dt_sel,
            dt_tell=self._dt_tell,
        )

    def _reset_timing(self) -> None:
        self._dt_fit, self._dt_gen, self._dt_sel = 0.0, 0.0, 0.0

    def ask(self, num_arms: int) -> np.ndarray:
        num_arms = int(num_arms)
        if num_arms <= 0:
            raise ValueError(num_arms)
        if self._tr_state is not None:
            self._tr_state.validate_request(num_arms)
        early = self._mode_impl.try_early_ask(
            num_arms,
            self._x_obs_list,
            self._draw_initial,
            self._get_init_lhd_points,
        )
        if early is not None:
            self._reset_timing()
            return early
        if self._init_idx < self._num_init:

            def fallback(n):
                return self._ask_normal(n, is_fallback=True)

            self._reset_timing()
            return self._get_init_lhd_points(
                num_arms, fallback_fn=fallback if self._x_obs_list else None
            )
        if not self._x_obs_list:
            self._reset_timing()
            return self._draw_initial(num_arms)
        return self._ask_normal(num_arms)

    def _handle_tr_restart(self, num_arms: int) -> np.ndarray | None:
        if not self._tr_state.needs_restart():
            return None
        self._tr_state.restart()
        should_reset_init, new_init_idx = self._call_handle_restart()
        if not should_reset_init:
            return None
        self._y_tr_list, self._init_idx = [], new_init_idx
        self._init_lhd = turbo_utils.from_unit(
            turbo_utils.latin_hypercube(self._num_init, self._num_dim, rng=self._rng),
            self._bounds,
        )
        return self._get_init_lhd_points(num_arms)

    def _call_handle_restart(self) -> tuple[bool, int]:
        from . import impl_helpers

        if self._mode_impl.always_clears_on_restart:
            return impl_helpers.handle_restart_clear_always(
                self._x_obs_list, self._y_obs_list, self._yvar_obs_list
            )
        return impl_helpers.handle_restart_check_morbo(
            self._config,
            self._x_obs_list,
            self._y_obs_list,
            self._yvar_obs_list,
            self._init_idx,
        )

    def _ask_normal(self, num_arms: int, *, is_fallback: bool = False) -> np.ndarray:
        import time

        if self._tr_state is None:
            return self._draw_initial(num_arms)
        restart_result = self._handle_tr_restart(num_arms)
        if restart_result is not None:
            return restart_result
        if self._mode_impl.needs_tr_list() and not self._x_obs_list:
            return self._get_init_lhd_points(num_arms)

        t0_fit = time.perf_counter()
        _, _, _, lengthscales = self._mode_impl.prepare_ask(
            self._x_obs_list,
            self._y_obs_list,
            self._yvar_obs_list,
            self._num_dim,
            self._gp_num_steps,
            rng=self._rng,
        )
        self._dt_fit = time.perf_counter() - t0_fit

        x_center = self._mode_impl.get_x_center(
            self._x_obs_list,
            self._y_obs_list,
            self._rng,
            self._tr_state,
        )
        if x_center is None:
            if not self._y_obs_list:
                raise RuntimeError("no observations")
            x_center = np.full(self._num_dim, 0.5)

        t0_gen = time.perf_counter()
        x_cand = self._generate_tr_candidates(
            x_center,
            lengthscales,
            self._num_candidates,
            self._rng,
            num_arms=num_arms,
        )
        self._dt_gen = time.perf_counter() - t0_gen

        def from_unit_fn(x):
            return turbo_utils.from_unit(x, self._bounds)

        def fallback_fn(x, n):
            from .proposal import select_uniform

            return select_uniform(x, n, self._num_dim, self._rng, from_unit_fn)

        self._tr_state.validate_request(num_arms, is_fallback=is_fallback)

        t0_sel = time.perf_counter()
        selected = self._mode_impl.select_candidates(
            x_cand,
            num_arms,
            self._num_dim,
            self._rng,
            fallback_fn,
            from_unit_fn,
            tr_state=self._tr_state,
        )
        self._dt_sel = time.perf_counter() - t0_sel
        return selected

    def _trim_trailing_obs(self) -> None:
        y_tr_array = np.asarray(self._y_tr_list, dtype=float)
        incumbent_indices = self._tr_state.get_incumbent_indices(y_tr_array, self._rng)
        self._x_obs_list, self._y_obs_list, self._y_tr_list, self._yvar_obs_list = (
            turbo_optimizer_utils.trim_trailing_observations(
                self._x_obs_list,
                self._y_obs_list,
                self._y_tr_list,
                self._yvar_obs_list,
                trailing_obs=self._trailing_obs,
                incumbent_indices=incumbent_indices,
            )
        )

    def tell(
        self, x: np.ndarray, y: np.ndarray, y_var: np.ndarray | None = None
    ) -> np.ndarray:
        with turbo_utils.record_duration(
            lambda dt: setattr(self, "_dt_tell", float(dt))
        ):
            x, y, y_var, num_metrics = turbo_optimizer_utils.validate_tell_inputs(
                x, y, y_var, self._num_dim
            )
            if self._config.tr_type != "morbo" and num_metrics != 1:
                raise ValueError(
                    f"Single-objective requires num_metrics=1, got {num_metrics}"
                )
            if self._tr_state is None:
                self._tr_state = self._create_trust_region(
                    self._num_dim, x.shape[0], self._rng, num_metrics=num_metrics
                )
            cfg_nm = self._config.num_metrics
            if cfg_nm is not None and num_metrics != cfg_nm:
                raise ValueError(f"y has {num_metrics} metrics but expected {cfg_nm}")
            if self._expects_yvar is None:
                self._expects_yvar = y_var is not None
            if (y_var is not None) != bool(self._expects_yvar):
                raise ValueError(
                    f"y_var must be {'provided' if self._expects_yvar else 'omitted'} on every tell()"
                )
            if x.shape[0] == 0:
                return (
                    np.array([], dtype=float)
                    if num_metrics == 1
                    else np.empty((0, num_metrics), dtype=float)
                )

            x_unit = turbo_utils.to_unit(x, self._bounds)
            self._x_obs_list.extend(x_unit.tolist())
            self._y_obs_list.extend(y.tolist())
            if y_var is not None:
                self._yvar_obs_list.extend(y_var.tolist())

            x_all, y_all = (
                np.asarray(self._x_obs_list, dtype=float),
                np.asarray(self._y_obs_list, dtype=float),
            )
            self._mode_impl.prepare_ask(
                self._x_obs_list,
                self._y_obs_list,
                self._yvar_obs_list,
                self._num_dim,
                0,
                rng=self._rng,
            )
            mu_all = np.asarray(self._mode_impl.estimate_y(x_all, y_all), dtype=float)
            y_estimate = np.asarray(self._mode_impl.estimate_y(x_unit, y), dtype=float)
            self._y_tr_list = mu_all.tolist()

            if self._trailing_obs is not None:
                self._trim_trailing_obs()

            prev_n = int(getattr(self._tr_state, "prev_num_obs", 0))
            if (
                prev_n > 0
                and prev_n <= len(self._y_tr_list)
                and hasattr(self._tr_state, "best_value")
            ):
                y_tr = np.asarray(self._y_tr_list, dtype=float)
                self._tr_state.best_value = float(
                    np.max((y_tr[:, 0] if y_tr.ndim == 2 else y_tr)[:prev_n])
                )

            self._tr_state.update(np.asarray(self._y_tr_list, dtype=float))
            return y_estimate

    def _create_trust_region(
        self, num_dim: int, num_arms: int, rng: Any, num_metrics: int | None = None
    ) -> Any:
        from . import impl_helpers

        return impl_helpers.create_trust_region(
            self._config, num_dim, num_arms, rng, num_metrics
        )

    def _generate_tr_candidates(
        self,
        x_center: np.ndarray,
        lengthscales: np.ndarray | None,
        num_candidates: int,
        rng: Any,
        *,
        num_arms: int,
        candidate_rv: str = "sobol",
    ) -> np.ndarray:
        from . import tr_helpers

        if candidate_rv == "sobol":
            from scipy.stats import qmc

            sobol_seed = turbo_optimizer_utils.sobol_seed_for_state(
                self._sobol_seed_base, n_obs=len(self._x_obs_list), num_arms=num_arms
            )
            sobol_engine = qmc.Sobol(d=self._num_dim, scramble=True, seed=sobol_seed)
        else:
            sobol_engine = None

        return tr_helpers.generate_tr_candidates(
            self._tr_state.compute_bounds_1d,
            x_center,
            lengthscales,
            num_candidates,
            rng=rng,
            candidate_rv=candidate_rv,
            sobol_engine=sobol_engine,
        )

    def _draw_initial(self, num_arms: int) -> np.ndarray:
        return turbo_utils.from_unit(
            turbo_utils.latin_hypercube(num_arms, self._num_dim, rng=self._rng),
            self._bounds,
        )

    def _get_init_lhd_points(
        self, num_arms: int, fallback_fn: Callable[[int], np.ndarray] | None = None
    ):
        num_to_return = min(num_arms, self._num_init - self._init_idx)
        result = self._init_lhd[self._init_idx : self._init_idx + num_to_return]
        self._init_idx += num_to_return
        if num_to_return < num_arms:
            extra = (fallback_fn or self._draw_initial)(num_arms - num_to_return)
            result = np.vstack([result, extra])
        return result



from __future__ import annotations

import warnings

import numpy as np
import pytest
import torch
from gpytorch.constraints import Interval
from gpytorch.distributions import MultivariateNormal
from gpytorch.likelihoods import FixedNoiseGaussianLikelihood
from gpytorch.mlls import ExactMarginalLogLikelihood

from enn.turbo.turbo_gp_fit import fit_gp
from enn.turbo.turbo_gp_noisy import TurboGPNoisy


def _make_turbo_gp_noisy(
    *,
    train_x,
    train_y,
    train_y_var,
    ard_dims: int,
    learn_additional_noise: bool = False,
):
    lengthscale_constraint = Interval(0.005, 2.0)
    outputscale_constraint = Interval(0.05, 20.0)
    return TurboGPNoisy(
        train_x=train_x,
        train_y=train_y,
        train_y_var=train_y_var,
        lengthscale_constraint=lengthscale_constraint,
        outputscale_constraint=outputscale_constraint,
        ard_dims=ard_dims,
        learn_additional_noise=learn_additional_noise,
    )


def test_fit_gp_returns_model_with_valid_data():
    num_obs, num_dim = 20, 3
    x_obs = np.random.default_rng(0).random((num_obs, num_dim))
    y_obs = (
        x_obs.sum(axis=1) + 0.1 * np.random.default_rng(1).standard_normal(num_obs)
    ).tolist()
    model, likelihood, y_mean, y_std = fit_gp(
        x_obs.tolist(), y_obs, num_dim, num_steps=10
    )
    assert model is not None and likelihood is not None
    assert isinstance(y_mean, float) and isinstance(y_std, float) and y_std > 0.0


def test_fit_gp_returns_none_with_empty_data_and_returns_model_with_single_obs():
    num_dim = 2
    model_empty, likelihood_empty, mean_empty, std_empty = fit_gp(
        [], [], num_dim, num_steps=10
    )
    assert model_empty is None and likelihood_empty is None
    assert mean_empty == 0.0 and std_empty == 1.0

    x_single = np.random.default_rng(0).random((1, num_dim))
    model_single, likelihood_single, mean_single, std_single = fit_gp(
        x_single.tolist(), [1.0], num_dim, num_steps=0
    )
    assert model_single is not None and likelihood_single is not None
    assert mean_single == 1.0 and std_single == 1.0


def test_fit_gp_with_y_var_list_uses_noisy_model():
    num_obs, num_dim = 20, 3
    rng = np.random.default_rng(42)
    x_obs = rng.random((num_obs, num_dim))
    y_obs = (x_obs.sum(axis=1) + 0.1 * rng.standard_normal(num_obs)).tolist()
    y_var = rng.uniform(0.01, 0.1, size=num_obs).tolist()
    model, likelihood, y_mean, y_std = fit_gp(
        x_obs.tolist(), y_obs, num_dim, yvar_obs_list=y_var, num_steps=10
    )
    assert model is not None and isinstance(model, TurboGPNoisy)
    assert likelihood is not None and y_std > 0.0


def test_fit_gp_with_y_var_list_asserts_length():
    num_obs, num_dim = 10, 2
    rng = np.random.default_rng(0)
    x_obs = rng.random((num_obs, num_dim)).tolist()
    y_obs = rng.random(num_obs).tolist()
    y_var_wrong = rng.uniform(0.01, 0.1, size=num_obs - 2).tolist()
    with pytest.raises(ValueError):
        fit_gp(x_obs, y_obs, num_dim, yvar_obs_list=y_var_wrong, num_steps=5)


def test_turbo_gp_noisy_accepts_train_y_var():
    num_obs, num_dim = 10, 2
    rng = np.random.default_rng(42)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(
        train_x.sum(dim=1).numpy() + 0.1 * rng.standard_normal(num_obs),
        dtype=torch.float64,
    )
    train_y_var = torch.as_tensor(
        rng.uniform(0.01, 0.1, size=num_obs), dtype=torch.float64
    )
    model = _make_turbo_gp_noisy(
        train_x=train_x, train_y=train_y, train_y_var=train_y_var, ard_dims=num_dim
    )
    assert model is not None and isinstance(
        model.likelihood, FixedNoiseGaussianLikelihood
    )


def test_turbo_gp_noisy_forward_and_posterior():
    num_obs, num_dim = 15, 3
    rng = np.random.default_rng(123)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(
        train_x.sum(dim=1).numpy() + 0.05 * rng.standard_normal(num_obs),
        dtype=torch.float64,
    )
    train_y_var = torch.full((num_obs,), 0.01, dtype=torch.float64)
    model = _make_turbo_gp_noisy(
        train_x=train_x, train_y=train_y, train_y_var=train_y_var, ard_dims=num_dim
    )
    model.eval()
    model.likelihood.eval()
    test_x = torch.as_tensor(rng.random((5, num_dim)), dtype=torch.float64)
    with torch.no_grad():
        forward_output = model.forward(test_x)
        posterior_output = model.posterior(test_x)
    assert isinstance(forward_output, MultivariateNormal) and isinstance(
        posterior_output, MultivariateNormal
    )
    assert forward_output.mean.shape == (5,) and posterior_output.mean.shape == (5,)


def test_turbo_gp_noisy_trains_successfully():
    num_obs, num_dim = 20, 2
    rng = np.random.default_rng(999)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(
        train_x.sum(dim=1).numpy() + 0.1 * rng.standard_normal(num_obs),
        dtype=torch.float64,
    )
    train_y_var = torch.as_tensor(
        rng.uniform(0.005, 0.05, size=num_obs), dtype=torch.float64
    )
    model = _make_turbo_gp_noisy(
        train_x=train_x, train_y=train_y, train_y_var=train_y_var, ard_dims=num_dim
    )
    model.train()
    model.likelihood.train()
    mll = ExactMarginalLogLikelihood(model.likelihood, model)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)
    initial_loss = None
    for i in range(20):
        optimizer.zero_grad()
        output = model(train_x)
        loss = -mll(output, train_y)
        if i == 0:
            initial_loss = loss.item()
        loss.backward()
        optimizer.step()
    assert loss.item() <= initial_loss


def test_turbo_gp_noisy_with_zero_variance():
    num_obs, num_dim = 10, 2
    rng = np.random.default_rng(42)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(train_x.sum(dim=1).numpy(), dtype=torch.float64)
    train_y_var = torch.zeros(num_obs, dtype=torch.float64)
    from gpytorch.utils.warnings import NumericalWarning

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"Very small noise values detected\..*",
            category=NumericalWarning,
        )
        model = _make_turbo_gp_noisy(
            train_x=train_x,
            train_y=train_y,
            train_y_var=train_y_var,
            ard_dims=num_dim,
            learn_additional_noise=True,
        )
    model.eval()
    model.likelihood.eval()
    test_x = torch.as_tensor(rng.random((3, num_dim)), dtype=torch.float64)
    with torch.no_grad():
        posterior = model.posterior(test_x)
    assert posterior.mean.shape == (3,)


def test_fit_gp_multi_output_can_trigger_non_scalar_backward_error():
    rng = np.random.default_rng(0)
    num_dim, num_metrics, n = 3, 2, 8
    x = rng.uniform(0.0, 1.0, size=(n, num_dim))
    y = rng.normal(size=(n, num_metrics))
    model, likelihood, _, _ = fit_gp(x.tolist(), y.tolist(), num_dim, num_steps=0)
    assert model is not None and likelihood is not None
    model.train()
    likelihood.train()
    train_x = model.train_inputs[0]
    train_y = model.train_targets
    output = model(train_x)
    mll = ExactMarginalLogLikelihood(likelihood, model)
    loss = -mll(output, train_y)
    assert tuple(loss.shape) == (num_metrics,)
    with pytest.raises(
        RuntimeError, match="grad can be implicitly created only for scalar outputs"
    ):
        loss.backward()


def test_fit_gp_multi_output_trains_without_scalar_backward_error():
    rng = np.random.default_rng(0)
    num_dim, num_metrics, n = 3, 2, 8
    x = rng.uniform(0.0, 1.0, size=(n, num_dim))
    y = rng.normal(size=(n, num_metrics))
    model, likelihood, gp_y_mean, gp_y_std = fit_gp(
        x.tolist(), y.tolist(), num_dim, num_steps=2
    )
    assert model is not None and likelihood is not None
    assert gp_y_mean.shape == (num_metrics,) and gp_y_std.shape == (num_metrics,)
